"""
MIT License
Copyright (c) 2016 Ionata Digital
Copyright (c) 2010-2014 Joshua Roesslein

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from __future__ import print_function

import json
import enum
from datetime import datetime

import six

from email.utils import parsedate


def parse_datetime(string):
    return datetime(*(parsedate(string)[:6]))


def parse_html_value(html):
    return html[html.find('>')+1:html.rfind('<')]


def parse_a_href(atag):
    start = atag.find('"') + 1
    end = atag.find('"', start)
    return atag[start:end]


def convert_to_utf8_str(arg):
    # written by Michael Norton (http://docondev.blogspot.com/)
    if isinstance(arg, six.text_type):
        arg = arg.encode('utf-8')
    elif not isinstance(arg, bytes):
        arg = six.text_type(arg).encode('utf-8')
    return arg


def import_simplejson():
    try:
        import simplejson as json
    except ImportError:
        try:
            import json  # Python 2.6+
        except ImportError:
            try:
                # Google App Engine
                from django.utils import simplejson as json
            except ImportError:
                raise ImportError("Can't load a json library.")

    return json


def list_to_csv(item_list):
    if item_list:
        return ','.join([str(i) for i in item_list])


class SenseTEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, enum.Enum):
            return obj.value
        return json.JSONEncoder.default(self, obj)



"""
MIT License
Copyright (c) 2016 Ionata Digital
Copyright (c) 2009-2014 Joshua Roesslein

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""
from __future__ import unicode_literals, absolute_import, print_function

from senaps_sensor.binder import bind_api
from senaps_sensor.error import SenapsError
from senaps_sensor.parsers import ModelParser, Parser
from senaps_sensor.utils import list_to_csv
from senaps_sensor.const import VALID_PROTOCOLS


class API(object):
    """Senaps API"""

    def __init__(self, auth_handler=None,
                 host='senaps.io', cache=None, api_root='/api/sensor/v2',
                 retry_count=0, retry_delay=0, retry_errors=None, timeout=60, parser=None,
                 compression=False, wait_on_rate_limit=False, connect_retries=3, read_retries=3,
                 backoff_factor=0.5, status_retries=3,
                 wait_on_rate_limit_notify=False, proxy='', verify=True, protocol='https'):
        """ Api instance Constructor

        :param auth_handler:
        :param host:  url of the server of the rest api, default:'senaps.io'
        :param cache: Cache to query if a GET method is used, default:None
        :param api_root: suffix of the api version, default:'/1.1'
        :param retry_count: number of allowed retries, default:0
        :param retry_delay: delay in second between retries, default:0
        :param retry_errors: default:None
        :param timeout: delay before to consider the request as timed out in seconds, default:60
        :param parser: ModelParser instance to parse the responses, default:None
        :param compression: If the response is compressed, default:False
        :param wait_on_rate_limit: If the api wait when it hits the rate limit, default:False
        :param wait_on_rate_limit_notify: If the api print a notification when the rate limit is hit, default:False
        :param proxy: Url to use as proxy during the HTTP request, default:''
        :param protocol: specify connection protocol to use. https by default.
        :param verify: Verify SSL certs if true. Will have no affect if protocol='http'
        :raise TypeError: If the given parser is not a ModelParser instance.
        :raise ValueError: If the given protocol is not in the set 'http', 'https'
        """
        self.auth = auth_handler
        self.protocol = protocol.lower()
        self.verify = verify
        self.host = host
        self.api_root = api_root
        self.cache = cache
        self.compression = compression
        self.retry_count = retry_count
        self.retry_delay = retry_delay
        self.retry_errors = retry_errors
        self.connect_retries = connect_retries
        self.read_retries = read_retries
        self.status_retries = status_retries
        self.backoff_factor = backoff_factor
        self.timeout = timeout
        self.wait_on_rate_limit = wait_on_rate_limit
        self.wait_on_rate_limit_notify = wait_on_rate_limit_notify
        self.parser = parser or ModelParser()
        self.proxy = {}

        if self.protocol not in VALID_PROTOCOLS:
            raise ValueError('"protocol" argument must be in %s' % (','.join(VALID_PROTOCOLS)))

        if proxy:
            self.proxy[self.protocol] = proxy

        parser_type = Parser
        if not isinstance(self.parser, parser_type):
            raise TypeError(
                '"parser" argument has to be an instance of "{required}".'
                ' It is currently a {actual}.'.format(
                    required=parser_type.__name__,
                    actual=type(self.parser)
                )
            )

    @property
    def me(self):
        """ Get the authenticated user using root api call
            :reference: https://senaps.io/api-docs/#/default/get_
            :allowed_param: 
        """
        res = bind_api(
            api=self,
            path='/',
            payload_type='json',
            allowed_param=[],
            require_auth=True,
        )
        return self.get_user(id=res()['_embedded']['user'][0]['id'])

    @property
    def users(self):
        """ :reference: https://senaps.io/api-docs/#!/default/get_users
            :allowed_param: 'id', 'roleids'
        """
        return bind_api(
            api=self,
            path='/users',
            payload_type='user',
            payload_list=True,
            query_only_param=[
                'id',
                'roleids'
            ],
            require_auth=True,
        )

    @property
    def get_user(self):
        """ :reference: https://senaps.io/api-docs/#!/default/get_users_userid
            :allowed_param: 'id'
        """
        return bind_api(
            api=self,
            path='/users/{id}',
            payload_type='user',
            allowed_param=['id'],
            require_auth=True,
        )

    @property
    def create_user(self):
        """ :reference: https://senaps.io/api-docs/#!/default/put_users_userid
            :allowed_param: 'id', 'hidden', 'roleids'
        """
        return bind_api(
            api=self,
            method='PUT',
            path='/users/{id}',
            payload_type='user',
            allowed_param=['id',
                           'hidden',
                           'roleids',
                           'eulaids'],
            require_auth=True,
        )

    @property
    def update_user(self):
        """ :reference: https://senaps.io/api-docs/#!/default/put_users_userid
            :allowed_param: 'id', 'hidden', 'roleids'
        """
        return bind_api(
            api=self,
            method='PUT',
            path='/users/{id}',
            payload_type='user',
            allowed_param=['id',
                           'hidden',
                           'roleids',
                           'eulaids'],
            require_auth=True,
        )

    @property
    def delete_user(self):
        """ :reference: https://senaps.io/api-docs/#!/default/delete_users_userid
            :allowed_param: 'id'
        """
        return bind_api(
            api=self,
            method='DELETE',
            path='/users/{id}',
            payload_type='user',
            allowed_param=['id'],
            require_auth=True,
        )

    @property
    def roles(self):
        """ :reference: https://senaps.io/api-docs/#!/default/get_roles
            :allowed_param: 'id', 'name', 'organisationid', 'groupids', 'streamids'
        """
        return bind_api(
            api=self,
            path='/roles',
            payload_type='role',
            payload_list=True,
            query_only_param=[
                'id',
                'type',
                'permissions',
                'organisationids',
                'groupids',
                'limit',
                'skip',
                'expand',
                'recursive',
                'usermetadatafield',
                'usermetadatavalues'
            ],
            require_auth=True,
        )

    @property
    def get_role(self):
        """ :reference: https://senaps.io/api-docs/#!/default/get_roles_roleid
            :allowed_param: 'id'
        """
        return bind_api(
            api=self,
            method='GET',
            path='/roles/{id}',
            payload_type='role',
            allowed_param=['id'],
            require_auth=True,
        )

    @property
    def create_role(self):
        """ :reference: https://senaps.io/api-docs/#!/default/put_roles_roleid
            :allowed_param: 'id'
        """
        return bind_api(
            api=self,
            method='PUT',
            path='/roles/{id}',
            payload_type='role',
            allowed_param=[
                'id',
                'permissions',
                'type',
                'organisationid',
                'groupid',
                'addressfilters',
                'usermetadata',
            ],
            require_auth=True,
        )

    @property
    def update_role(self):
        """ :reference: https://senaps.io/api-docs/#!/default/put_roles_roleid
            :allowed_param: 'id'
        """
        return bind_api(
            api=self,
            method='PUT',
            path='/roles/{id}',
            payload_type='role',
            allowed_param=[
                'id',
                'permissions',
                'type',
                'organisationid',
                'groupid',
                'addressfilters',
                'usermetadata',
            ],
            require_auth=True,
        )

    @property
    def delete_role(self):
        """ :reference: https://senaps.io/api-docs/#!/default/delete_roles_roleid
            :allowed_param: 'id'
        """
        return bind_api(
            api=self,
            method='DELETE',
            path='/roles/{id}',
            payload_type='json',
            allowed_param=['id'],
            require_auth=True,
        )

    @property
    def platforms(self):
        """ :reference: https://senaps.io/api-docs/#/default/platforms
            :allowed_param: 'id', 'name', 'organisationid', 'groupids', 'streamids'
        """
        return bind_api(
            api=self,
            path='/platforms',
            payload_type='platform',
            payload_list=True,
            query_only_param=[
                'id',
                'name',
                'organisationid',
                'groupids',
                'streamids',
                'usermetadatafield',
                'usermetadatavalues',
                'expand',
                'recursive',
            ],
            require_auth=True,
        )

    @property
    def create_platform(self):
        """ :reference: https://senaps.io/api-docs/#/default/put_platforms_id
            :allowed_param: 'id', 'name', 'organisationid', 'groupids', 'streamids', 'deployments'
        """
        return bind_api(
            api=self,
            path='/platforms/{id}',
            method='PUT',
            payload_type='platform',
            action='create',
            allowed_param=[
                'id',
                'name',
                'organisationid',
                'groupids',
                'streamids',
                'deployments',
                'usermetadata',
            ],
            require_auth=True,
        )

    @property
    def update_platform(self):
        """ :reference: https://senaps.io/api-docs/#/default/put_platforms_id
            :allowed_param: 'id', 'name', 'organisationid', 'groupids', 'streamids', 'deployments',
        """
        return bind_api(
            api=self,
            path='/platforms/{id}',
            method='PUT',
            payload_type='platform',
            action='update',
            allowed_param=[
                'id',
                'name',
                'organisationid',
                'groupids',
                'streamids',
                'deployments',
                'usermetadata',
            ],
            require_auth=True,
        )

    @property
    def destroy_platform(self):
        """ :reference: https://senaps.io/api-docs/#/default/delete_platforms_id
            :allowed_param: 'id', 'cascade'
        """
        return bind_api(
            api=self,
            path='/platforms/{id}',
            method='DELETE',
            payload_type='platform',
            allowed_param=[
                'id',
            ],
            query_only_param=[
                'cascade',
            ],
            require_auth=True,
        )

    @property
    def get_platform(self):
        """ :reference: https://senaps.io/api-docs/#/default/get_platform_id
            :allowed_param: 'id'
        """
        return bind_api(
            api=self,
            path='/platforms/{id}',
            method='GET',
            payload_type='platform',
            allowed_param=['id'],
            require_auth=True,
        )

    @property
    def streams(self):
        """ :reference: https://senaps.io/api-docs/#!/default/get_streams
            :allowed_param: 'id,limit'
        """
        return bind_api(
            api=self,
            path='/streams',
            payload_type='stream',
            allowed_param=['limit', 'id'],
            query_only_param=[
                'id',
                'limit',
                'skip',
                'near',
                'radius',
                'resulttype',
                'expand',
                'recursive',
                'groupids',
                'organisationid',
                'locationid',
                'usermetadatafield',
                'usermetadatavalues',
                'properties'
            ],
            payload_list=True,
            require_auth=True,
        )

    @property
    def locations(self):
        """ :reference: https://senaps.io/api-docs/#!/default/get_locations
            :allowed_param: 'id,limit'
        """
        return bind_api(
            api=self,
            path='/locations',
            payload_type='location',
            allowed_param=['id'],
            query_only_param=[
                'id',
                'description',
                'limit',
                'skip',
                'near',
                'radius',
                'expand',
                'groupids',
                'organisationid',
                'usermetadatafield',
                'usermetadatavalues',
            ],
            payload_list=True,
            require_auth=True,
        )

    @property
    def get_stream(self):
        """ :reference: https://senaps.io/api-docs/#!/default/get_streams_id
            :allowed_param: 'id'
        """
        return bind_api(
            api=self,
            path='/streams/{id}',
            method='GET',
            payload_type='stream',
            allowed_param=['id'],
            require_auth=True,
        )

    @property
    def create_stream(self):
        """ :reference: https://senaps.io/api-docs/#!/default/put_streams_id
            :allowed_param: 'id', 'resulttype', 'organisationid', 'groupids', 'procedureid', 'samplePeriod',
            'reportingPeriod', 'streamMetadata', 'locationid'
        """
        return bind_api(
            api=self,
            path='/streams/{id}',
            method='PUT',
            payload_type='stream',
            action='create',
            allowed_param=[
                'id',
                'resulttype',
                'organisationid',
                'groupids',
                'locationid',
                'procedureid',
                'samplePeriod',
                'reportingPeriod',
                'streamMetadata',
                'usermetadata',
            ],
            require_auth=True,
        )

    @property
    def update_stream(self):
        return self.create_stream

    @property
    def destroy_stream(self):
        """ :reference: https://senaps.io/api-docs/#!/default/delete_streams_id
            :allowed_param: 'id', 'cascade'
        """
        return bind_api(
            api=self,
            path='/streams/{id}',
            method='DELETE',
            payload_type='stream',
            allowed_param=[
                'id',
            ],
            query_only_param=[
                'cascade',
            ],
            require_auth=True,
        )

    @property
    def create_location(self):
        """ :reference: https://senaps.io/api-docs/#!/default/put_location_id
            :allowed_param: 'id', 'description', 'organisationid', 'geoJson'
        """
        return bind_api(
            api=self,
            path='/locations/{id}',
            method='PUT',
            payload_type='json',
            action='create',
            allowed_param=[
                'id',
                'organisationid',
                'description',
                'geoJson',
                'groupids',
                'usermetadata',
            ],
            require_auth=True,
        )

    @property
    def get_location(self):
        """ :reference: https://senaps.io/api-docs/#!/default/get_location_id
            :allowed_param: 'id'
        """
        return bind_api(
            api=self,
            path='/locations/{id}',
            method='GET',
            payload_type='location',
            allowed_param=['id'],
            require_auth=True,
        )

    @property
    def destroy_location(self):
        """ :reference: https://senaps.io/api-docs/#!/default/delete_locations_id
            :allowed_param: 'id', 'cascade'
        """
        return bind_api(
            api=self,
            path='/locations/{id}',
            method='DELETE',
            payload_type='location',
            allowed_param=[
                'id',
            ],
            query_only_param=[
                'cascade',
            ],
            require_auth=True,
        )

    @property
    def get_aggregation(self):
        """ :reference: https://senaps.io/api-docs/#!/default/get_aggregation
            :allowed_param: 'streamid', 'start', 'end', 'si', 'ei',
            'limit', 'aggperiod', 'count'
        """
        return bind_api(
            api=self,
            path='/aggregation',
            method='GET',
            payload_type='json',
            allowed_param=[
            ],
            query_only_param=[
                'streamid',
                'start',
                'end',
                'si',
                'ei',
                'limit',
                'aggperiod',
                'count',
            ],
            require_auth=True,
        )

    @property
    def create_observations(self):
        """ :reference: https://senaps.io/api-docs/#!/default/post_observations
            :allowed_param: 'streamid', 'results'
        """
        return bind_api(
            api=self,
            path='/observations',
            method='POST',
            payload_type='json',
            action='create',
            allowed_param=[
                'streamid',
                'results',
            ],
            query_only_param=[
                'streamid',
            ],
            require_auth=True,
        )

    @property
    def get_observations(self):
        """ :reference: https://senaps.io/api-docs/#!/default/get_observations
            :allowed_param: 'streamid', 'start', 'end', 'time', 'si', 'ei',
            'bounds', 'media', 'limit', 'sort'
        """
        return bind_api(
            api=self,
            path='/observations',
            method='GET',
            payload_type='json',
            allowed_param=[
            ],
            query_only_param=[
                'streamid',
                'start',
                'end',
                'time',
                'si',
                'ei',
                'bounds',
                'media',
                'limit',
                'sort',
            ],
            require_auth=True,
        )

    @property
    def destroy_observations(self):
        """ :reference: https://senaps.io/api-docs/#!/default/delete_observations
            :allowed_param: 'streamid'
        """
        return bind_api(
            api=self,
            path='/observations',
            method='DELETE',
            payload_type='json',
            query_only_param=[
                'streamid',
            ],
            require_auth=True,
        )

    @property
    def create_group(self):
        """ :reference:
        https://senaps.io/api-docs/#!/default/put_groups_id
            :allowed_param: 'id', 'name', 'organisationid', 'description', 'groupids'
        """
        return bind_api(
            api=self,
            path='/groups/{id}',
            method='PUT',
            payload_type='json',
            action='create',
            allowed_param=[
                'id',
                'name',
                'organisationid',
                'description',
                'groupids',
                'usermetadata',
            ],
            require_auth=True,
        )

    @property
    def get_groups(self):
        """ :reference: https://senaps.io/api-docs/#!/default/get_groups
            :allowed_param: 'id', 'organisationid', 'groupids', 'limit', 'skip', 'expand', 'recursive'
        """
        return bind_api(
            api=self,
            path='/groups',
            method='GET',
            payload_type='group',
            payload_list=True,
            allowed_param=[
            ],
            query_only_param=[
                'id',
                'organisationid',
                'groupids',
                'limit',
                'skip',
                'expand',
                'recursive',
                'usermetadatafield',
                'usermetadatavalues',
            ],
            require_auth=True,
        )

    @property
    def destroy_group(self):
        """ :reference: https://senaps.io/api-docs/#!/default/delete_group
            :allowed_param: 'id', 'cascade'
        """
        return bind_api(
            api=self,
            path='/groups/{id}',
            method='DELETE',
            payload_type='group',
            allowed_param=[
                'id',
            ],
            query_only_param=[
                'cascade',
            ],
            require_auth=True,
        )

    @property
    def get_group(self):
        """ :reference: https://senaps.io/api-docs/#!/default/get_group
            :allowed_param: 'id', 'recursive'
        """
        return bind_api(
            api=self,
            path='/groups/{id}',
            method='GET',
            payload_type='json',
            allowed_param=['id', 'recursive'],
            require_auth=True,
        )

    @property
    def get_permitted(self):
        """
        Currently no public documentation for this item.
        This endpoint is used to verify whether the current user is permitted to
        access the specified resources.
        :allowed_param: 'permission', 'resourceid', 'organisationid', 'groupids'
        :return:
        """
        return bind_api(
            api=self,
            path='/permitted',
            method='GET',
            payload_type='permitted',
            allowed_param=[],
            query_only_param=[
                'permission',
                'resourceid',
                'organisationid',
                'groupids'
            ],
            require_auth=True
        )



# -*- coding: utf-8 -*-

from __future__ import absolute_import, print_function

from requests import PreparedRequest
from requests.structures import CaseInsensitiveDict

from .config import SensorApiInternalTestCase, username, tape

import unittest
import six
if six.PY3:
    from unittest.case import skip
else:
    from unittest2.case import skip

class TestInternalAuth(SensorApiInternalTestCase):

    def test_consumer_header_exists(self):
        p = PreparedRequest()
        p.headers = CaseInsensitiveDict()
        r = self.api.auth(p)

        expected_header = 'X-Consumer-Custom-ID'
        self.assertIn(expected_header,
                      r.headers.keys(),
                      'Did not find %s in headers, auth must be broken' % expected_header)
        self.assertEqual(username,
                         r.headers['X-Consumer-Custom-ID'],
                         'Header not populated with username correctly.')



"""
MIT License
Copyright (c) 2016 Ionata Digital
Copyright (c) 2009-2014 Joshua Roesslein

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from __future__ import print_function

import six


class SenapsError(Exception):
    """Senaps exception"""

    def __init__(self, reason, response=None, api_code=None):
        self.reason = six.text_type(reason)
        self.response = response
        self.api_code = api_code
        Exception.__init__(self, reason)

    def __str__(self):
        return self.reason


def is_rate_limit_error_message(message):
    """Check if the supplied error message belongs to a rate limit error."""
    return isinstance(message, list) \
           and len(message) > 0 \
           and 'code' in message[0] \
           and message[0]['code'] == 88


class RateLimitError(SenapsError):
    """Exception for Senaps hitting the rate limit."""
    # RateLimitError has the exact same properties and inner workings
    # as SenseTError for backwards compatibility reasons.
    pass



"""
MIT License
Copyright (c) 2016 Ionata Digital

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""
from __future__ import unicode_literals, absolute_import, print_function

import json
import time
import datetime
import uuid

from senaps_sensor.error import SenapsError
from senaps_sensor.models import Deployment, Organisation, Group, Platform, Stream, StreamResultType, StreamMetaData, \
    StreamMetaDataType, StreamMetaDataMimeType, \
    InterpolationType, Observation, UnivariateResult, Location, User, Role

from senaps_sensor.const import VALID_PROTOCOLS

from senaps_sensor.utils import SenseTEncoder
from senaps_sensor.binder import bind_api

from tests.config import *

import six

if six.PY3:
    import unittest
    from unittest.case import skip
else:
    import unittest2 as unittest
    from unittest2.case import skip


def dumps(*args, **kwargs):
    if 'cls' not in kwargs:
        kwargs['cls'] = SenseTEncoder
    return json.dumps(*args, **kwargs)


class ApiTestCase(SensorApiTestCase):

    def setUp(self):
        super(ApiTestCase, self).setUp()

    def generate_location(self):
        o = Organisation()
        o.id = "sandbox"

        loc = Location()
        loc.organisations = [o]
        loc.id = str(uuid.uuid1())
        loc.geoJson = {'type': 'Point', 'coordinates': [147.0, -42.0]}

        return loc

    def generate_platform(self, location):
        o = Organisation()
        o.id = "sandbox"

        d = Deployment()
        d.name = 'Deployment 1'
        if location:
            d.location = location
        d.validTime = {}

        p = Platform()
        p.id = str(uuid.uuid1())
        p.name = "A Platform created for unittests"
        p.organisations = [o]
        p.deployments = [d]

        return p

    def generate_geolocation_stream(self, stream_id=None):
        sm = StreamMetaData()
        sm.type = StreamMetaDataType.geolocation
        sm.interpolation_type = InterpolationType.continuous
        s = self._generate_stream(StreamResultType.geolocation, sm, stream_id)

        return s

    def generate_scalar_stream(self, stream_id=None):
        sm = StreamMetaData()
        sm.type = StreamMetaDataType.scalar
        sm.interpolation_type = InterpolationType.continuous
        sm.observed_property = "http://registry.it.csiro.au/def/environment/property/air_temperature"
        sm.unit_of_measure = "http://registry.it.csiro.au/def/qudt/1.1/qudt-unit/DegreeCelsius"
        s = self._generate_stream(StreamResultType.scalar, sm, stream_id)
        return s

    def generate_vector_stream(self, length, stream_id=None):
        sm = StreamMetaData()
        sm.type = StreamMetaDataType.vector
        sm.length = length
        s = self._generate_stream(StreamResultType.vector, sm, stream_id)
        return s

    def generate_image_stream(self, stream_id=None):
        sm = StreamMetaData()
        sm.type = StreamMetaDataType.image

        s = self._generate_stream(StreamResultType.image, sm, stream_id)
        return s

    def generate_regularly_binned_vector_stream(self, start, end, step, stream_id=None):
        sm = StreamMetaData()
        sm.type = StreamMetaDataType.regularly_binned_vector
        sm.start = start
        sm.end = end
        sm.step = step

        sm.observed_property = "http://registry.it.csiro.au/def/environment/property/absorption_total"
        sm.amplitude_unit = "http://registry.it.csiro.au/def/qudt/1.1/qudt-unit/Percent"
        sm.length_unit = "http://registry.it.csiro.au/def/qudt/1.1/qudt-unit/Angstrom"

        s = self._generate_stream(StreamResultType.vector, sm, stream_id)
        return s

    def generate_document_stream(self, stream_id=None, mimetype=StreamMetaDataMimeType.json):
        sm = StreamMetaData()
        sm.type = StreamMetaDataType.document
        sm.mimetype = mimetype
        s = self._generate_stream(StreamResultType.document, sm, stream_id)
        return s

    def generate_group(self, id):

        g = Group()
        g.id = id
        g.name = 'Unit Test Group'

        return g

    def _generate_stream(self, stream_type, stream_meta_data, stream_id=None):

        o = Organisation()
        o.id = 'sandbox'

        s = Stream()
        s.id = stream_id if stream_id is not None else str(uuid.uuid1())

        s.organisations = [o]

        s.result_type = stream_type

        s.samplePeriod = 'PT10S'
        s.reportingPeriod = 'P1D'

        s.metadata = stream_meta_data

        return s

    def test_ensure_data_exists_in_senaps(self):
        # The following group, location must exist for future tests to pass
        # but we still assume the organisation integration_test exists
        test_group = self.api.create_group(id='integration_test',
                              name='Integration Test Group',
                              organisationid='integration_test')
        self.assertEqual(test_group.get('id'), 'integration_test')

        test_location = self.api.create_location(id='integration_test',
                                 organisationid='integration_test',
                                 description='Integration Test Group',
                                 geoJson={'type': 'Point', 'coordinates': [147.0, -42.0]},
                                 groupids=['integration_test'])
        self.assertEqual(test_location.get('id'), 'integration_test')

    @tape.use_cassette('test_create_platform.json')
    def test_create_platform(self):
        """
        Platform creation test, no clean up
        :return: None
        """
        # create
        loc = self.generate_location()
        p = self.generate_platform(loc)
        s = self.generate_scalar_stream()

        p.streams.append(s)

        print(loc.to_json("create"))

        actual_json = p.to_json("create")

        print(actual_json)

        # verify json
        required_json = dumps({
            "id": p.id,
            "name": p.name,
            "organisationid": p.organisations[0].id,
            "groupids": [

            ],
            "streamids": [
                s.id
            ],
            "deployments": [
                {"name": "Deployment 1", "locationid": loc.id, "validTime": {}}
            ]
        }, sort_keys=True)  # be explict with key order since dumps gives us a string

        self.assertEqual(actual_json, required_json)

        self.api.create_stream(s)
        self.api.create_location(loc)
        self.api.create_platform(p)

        created_platform = self.api.get_platform(id=p.id)

        print(created_platform)

        # verify
        self.assertEqual(created_platform.id, p.id)
        self.assertEqual(created_platform.name, p.name)
        self.assertEqual(created_platform.deployments[0].location.id, loc.id)
        self.assertEqual(created_platform.streams[0].id, s.id)

        location = self.api.get_location(id=loc.id)
        self.assertEqual(location.geojson['coordinates'][0], loc.geoJson['coordinates'][0])

        self.api.destroy_platform(id=p.id, cascade=True)
        self.api.destroy_location(id=loc.id, cascade=True)
        self.api.destroy_stream(id=s.id, cascade=True)

    def test_update_platform(self):
        """
        Platform update test, no clean ups
        :return: None
        """
        # create
        loc = self.generate_location()
        p = self.generate_platform(location=loc)
        self.api.create_location(loc)
        created_platform = self.api.create_platform(p)

        # update, by appending id to name attr
        created_platform.name += 'UPDATED'
        updated_platform = self.api.update_platform(created_platform)

        # verify
        self.assertEqual(updated_platform.name, created_platform.name)

    def test_update_platform_deployment_locationless(self):
        """
        Platform update test, no clean ups
        :return: None
        """
        # create
        p = self.generate_platform(location=None)
        created_platform = self.api.create_platform(p)

        # update, by appending id to name attr
        created_platform.name += 'UPDATED'
        updated_platform = self.api.update_platform(created_platform)

        # verify
        self.assertEqual(updated_platform.name, created_platform.name)

    def test_delete_platform(self):
        """
        Platform deletion test, create and cleanup
        :return: None
        """
        # create
        loc = self.generate_location()
        p = self.generate_platform(loc)
        self.api.create_location(loc)
        created_platform = self.api.create_platform(p)

        # delete
        self.api.destroy_platform(created_platform, cascade=True)

        # verify
        with self.assertRaises(SenapsError):
            self.api.get_platform(id=p.id)

        self.api.destroy_location(loc.id)

    @tape.use_cassette('test_non_existent_stream.json')
    def test_non_existent_stream(self):
        stream_nonexistent_id = str(uuid.uuid1())
        # stream_exists_id = "{0}_location".format(self.existing_platform_id)

        try:
            s = self.api.get_stream(id=stream_nonexistent_id)
        except SenapsError as ex:
            self.assertEqual(ex.api_code, 404)

    @tape.use_cassette('test_create_geolocation_stream.json')
    def test_create_geolocation_stream(self):
        s = self.generate_geolocation_stream()

        required_state = {
            "id": s.id,
            "resulttype": "geolocationvalue",
            "organisationid": s.organisations[0].id,
            "samplePeriod": "PT10S",
            "reportingPeriod": 'P1D',
            "streamMetadata": {
                "type": ".GeoLocationStreamMetaData",
                "interpolationType": "http://www.opengis.net/def/waterml/2.0/interpolationType/Continuous",
            }
        }
        required_json = dumps(required_state, sort_keys=True)  # be explict with key order since dumps gives us a string

        actual_state = s.to_state("create")
        actual_json = s.to_json("create")

        # dict diff
        # from deepdiff import DeepDiff
        # diff = DeepDiff(required_state, actual_state)

        # verify json
        self.assertEqual(actual_json, required_json)

        created_stream = self.api.create_stream(s)

        self.assertEqual(s.id, created_stream.id)

        # verify json
        self.assertEqual(actual_json, required_json)

        self.api.destroy_stream(id=s.id)

    def test_update_stream(self):

        s = self.generate_scalar_stream()
        s.samplePeriod = 'PT10S'

        self.api.create_stream(s)
        s.samplePeriod = 'PT20S'

        self.api.update_stream(s)
        updated_stream = self.api.get_stream(id=s.id)

        self.assertEqual(s.samplePeriod, updated_stream.samplePeriod)

        self.api.destroy_stream(id=s.id)

    def test_model_equals(self):

        loc1 = self.generate_location()
        loc2 = self.generate_location()

        self.assertNotEquals(loc1, loc2)

        # make location with same id as loc1
        o = Organisation()
        o.id = "sandbox"
        loc3 = Location()
        loc3.organisations = [o]
        loc3.id = loc1.id
        loc3.geoJson = {'type': 'Point', 'coordinates': [147.0, -42.0]}

        self.assertEqual(loc1, loc3)

    def test_update_stream_location(self):

        s = self.generate_scalar_stream()

        self.api.create_stream(s)

        retrieved_stream = self.api.get_stream(id=s.id)

        l = self.generate_location()

        self.api.create_location(l)

        retrieved_stream.location = l

        # update stream
        self.api.update_stream(retrieved_stream)

        updated_stream = self.api.get_stream(id=s.id)

        self.assertEqual(updated_stream.location.id, l.id)

        self.api.destroy_stream(id=s.id, cascade=True)

    def test_update_stream_with_results(self):

        s = self.generate_scalar_stream()

        self.api.create_stream(s)
        o, points = self.generate_observations()

        created_observation = self.api.create_observations(o, streamid=s.id)

        self.assertEqual(created_observation.get('message'), "Observations uploaded")
        self.assertEqual(created_observation.get('status'), 201)

        # retrieve stream (will now have resultsSummary)
        retrieved_stream = self.api.get_stream(id=s.id)
        # mutate stream a little
        retrieved_stream.samplePeriod = 'PT20S'
        # update stream
        self.api.update_stream(retrieved_stream)

        updated_stream = self.api.get_stream(id=s.id)

        self.assertEqual(updated_stream.samplePeriod, retrieved_stream.samplePeriod)

        self.api.destroy_observations(streamid=s.id)
        self.api.destroy_stream(id=s.id)

    def test_get_stream_enums(self):

        s = self.generate_scalar_stream()

        self.api.create_stream(s)

        retrieved_stream = self.api.get_stream(id=s.id)

        self.assertEqual(type(s.metadata.type), type(retrieved_stream.metadata.type))

        self.api.destroy_stream(id=s.id)

    @tape.use_cassette('test_create_scalar_stream.json')
    def test_create_scalar_stream(self):
        s = self.generate_scalar_stream()

        required_state = {
            "id": s.id,
            "resulttype": "scalarvalue",
            "organisationid": s.organisations[0].id,
            "samplePeriod": "PT10S",
            "reportingPeriod": 'P1D',
            "streamMetadata": {
                "type": ".ScalarStreamMetaData",
                "interpolationType": "http://www.opengis.net/def/waterml/2.0/interpolationType/Continuous",
                "observedProperty": "http://registry.it.csiro.au/def/environment/property/air_temperature",
                "unitOfMeasure": "http://registry.it.csiro.au/def/qudt/1.1/qudt-unit/DegreeCelsius"
            }
        }
        required_json = dumps(required_state, sort_keys=True)  # be explict with key order since dumps gives us a string

        actual_state = s.to_state("create")
        actual_json = s.to_json("create")

        # dict diff
        # from deepdiff import DeepDiff
        # diff = DeepDiff(required_state, actual_state)

        # verify json
        self.assertEqual(actual_json, required_json)

        created_stream = self.api.create_stream(s)

        self.assertEqual(s.id, created_stream.id)

        # verify json
        self.assertEqual(actual_json, required_json)

        self.api.destroy_stream(id=s.id)

    @tape.use_cassette('test_create_vector_stream.json')
    def test_create_vector_stream(self):
        s = self.generate_vector_stream(3)

        required_state = {
            "id": s.id,
            "resulttype": "vectorvalue",
            "organisationid": s.organisations[0].id,

            "samplePeriod": "PT10S",
            "reportingPeriod": 'P1D',
            "streamMetadata": {
                "type": ".VectorStreamMetaData",
                "length": 3
            }
        }
        required_json = dumps(required_state, sort_keys=True)  # be explict with key order since dumps gives us a string

        actual_state = s.to_state("create")
        actual_json = s.to_json("create")

        # dict diff
        # from deepdiff import DeepDiff
        # diff = DeepDiff(required_state, actual_state)

        # verify json
        self.assertEqual(actual_json, required_json)

        created_stream = self.api.create_stream(s)

        self.assertEqual(s.id, created_stream.id)

        # verify json
        self.assertEqual(actual_json, required_json)

        self.api.destroy_stream(id=s.id)

    @tape.use_cassette('test_create_regularly_binned_vector_stream.json')
    def test_create_regularly_binned_vector_stream(self):
        s = self.generate_regularly_binned_vector_stream(10, 20, 2, "ba94aada-84d0-420c-87d9-5510a17c176d")

        required_state = {
            "id": s.id,
            "resulttype": "vectorvalue",
            "organisationid": s.organisations[0].id,
            "samplePeriod": "PT10S",
            "reportingPeriod": 'P1D',
            "streamMetadata": {
                "type": ".RegularlyBinnedVectorStreamMetaData",
                "start": 10,
                "end": 20,
                "step": 2,
                "observedProperty": "http://registry.it.csiro.au/def/environment/property/absorption_total",
                "amplitudeUnit": "http://registry.it.csiro.au/def/qudt/1.1/qudt-unit/Percent",
                "lengthUnit": "http://registry.it.csiro.au/def/qudt/1.1/qudt-unit/Angstrom"
            }
        }
        required_json = dumps(required_state, sort_keys=True)  # be explict with key order since dumps gives us a string

        actual_state = s.to_state("create")
        actual_json = s.to_json("create")

        # dict diff
        # from deepdiff import DeepDiff
        # diff = DeepDiff(required_state, actual_state)

        # verify json
        self.assertEqual(actual_json, required_json)

        created_stream = self.api.create_stream(s)

        self.assertEqual(s.id, created_stream.id)

        # verify json
        self.assertEqual(actual_json, required_json)

        self.api.destroy_stream(id=s.id)

    @tape.use_cassette('test_create_document_json_stream.json')
    def test_create_document_json_stream(self):
        s = self.generate_document_stream()

        required_state = {
            "id": s.id,
            "resulttype": "documentvalue",
            "organisationid": s.organisations[0].id,
            "reportingPeriod": 'P1D',
            "samplePeriod": 'PT10S',
            "streamMetadata": {
                "type": ".DocumentStreamMetaData",
                "mimetype": "application/json"
            }
        }
        required_json = dumps(required_state, sort_keys=True)  # be explict with key order since dumps gives us a string

        actual_state = s.to_state("create")
        actual_json = s.to_json("create")

        # dict diff
        # from deepdiff import DeepDiff
        # diff = DeepDiff(required_state, actual_state)

        # verify json
        self.assertEqual(actual_json, required_json)

        created_stream = self.api.create_stream(s)

        self.assertEqual(s.id, created_stream.id)

        # verify json
        self.assertEqual(actual_json, required_json)

        self.api.destroy_stream(id=s.id)

    @tape.use_cassette('test_create_document_json_stream.json')
    def test_create_document_text_stream(self):
        s = self.generate_document_stream(mimetype=StreamMetaDataMimeType.text)

        required_state = {
            "id": s.id,
            "resulttype": "documentvalue",
            "organisationid": s.organisations[0].id,
            "reportingPeriod": 'P1D',
            "samplePeriod": 'PT10S',
            "streamMetadata": {
                "type": ".DocumentStreamMetaData",
                "mimetype": "text/plain"
            }
        }
        required_json = dumps(required_state, sort_keys=True)  # be explict with key order since dumps gives us a string

        actual_state = s.to_state("create")
        actual_json = s.to_json("create")

        # dict diff
        # from deepdiff import DeepDiff
        # diff = DeepDiff(required_state, actual_state)

        # verify json
        self.assertEqual(actual_json, required_json)

        created_stream = self.api.create_stream(s)

        self.assertEqual(s.id, created_stream.id)

        # verify json
        self.assertEqual(actual_json, required_json)

        self.api.destroy_stream(id=s.id)


    @tape.use_cassette('test_create_image_stream.json')
    def test_create_image_stream(self):
        s = self.generate_image_stream("403e2a68-7e4c-43e3-93d4-71d8980014fa")

        required_state = {
            "id": s.id,
            "resulttype": "imagevalue",
            "organisationid": s.organisations[0].id,
            "samplePeriod": "PT10S",
            "reportingPeriod": 'P1D',
            "streamMetadata": {
                "type": ".ImageStreamMetaData",
            }
        }
        required_json = dumps(required_state, sort_keys=True)  # be explict with key order since dumps gives us a string

        actual_state = s.to_state("create")
        actual_json = s.to_json("create")

        # dict diff
        # from deepdiff import DeepDiff
        # diff = DeepDiff(required_state, actual_state)

        # verify json
        self.assertEqual(actual_json, required_json)

        created_stream = self.api.create_stream(s)

        self.assertEqual(s.id, created_stream.id)

        # verify json
        self.assertEqual(actual_json, required_json)

        self.api.destroy_stream(id=s.id)

    @tape.use_cassette('test_create_geolocation_observations.json')
    def test_create_geolocation_observations(self):
        s = self.generate_geolocation_stream()

        o = Observation()

        dt = datetime.datetime(2016, 2, 15, 0, 0, 0)
        dt_td = datetime.timedelta(minutes=15)

        points = [
            {'time': dt + (dt_td * 0), "lng": 147.326262, "lat": -42.8840887, 'alt': 50},
            {'time': dt + (dt_td * 1), "lng": 147.3263529, "lat": -42.8844541},  # altitude missing, just because.
            {'time': dt + (dt_td * 2), "lng": 147.3232176, "lat": -42.883477, 'alt': 250},
        ]

        for p in points:
            item = UnivariateResult()
            item.t = p.get('time')
            coords = [p.get('lng'), p.get('lat'), p.get('alt')] if p.get('alt') is not None else [p.get('lng'),
                                                                                                  p.get('lat')]

            item.v = {
                'p': {
                    'type': 'Point',
                    'coordinates': coords
                }
            }
            o.results.append(item)

        dt_format = '%Y-%m-%dT%H:%M:%S.%fZ'
        required_state = {
            'results': [
                {
                    't': points[0].get('time').strftime(dt_format),
                    'v': {
                        'p': {
                            'type': 'Point',
                            'coordinates': [points[0].get('lng'), points[0].get('lat'), points[0].get('alt')]
                        }
                    }
                },
                {
                    't': points[1].get('time').strftime(dt_format),
                    'v': {
                        'p': {
                            'type': 'Point',
                            'coordinates': [points[1].get('lng'), points[1].get('lat')]
                        }
                    }
                },
                {
                    't': points[2].get('time').strftime(dt_format),
                    'v': {
                        'p': {
                            'type': 'Point',
                            'coordinates': [points[2].get('lng'), points[2].get('lat'), points[2].get('alt')]
                        }
                    }
                },
            ]
        }
        required_json = dumps(required_state, sort_keys=True)  # be explict with key order since dumps gives us a string

        actual_state = o.to_state("create")
        actual_json = o.to_json("create")

        ### dict diff debugging
        # from deepdiff import DeepDiff
        # diff = DeepDiff(required_state, actual_state)

        # verify json
        self.assertEqual(actual_json, required_json)

        created_stream = self.api.create_stream(s)

        self.assertEqual(created_stream.id, s.id)

        created_observation = self.api.create_observations(o, streamid=s.id)

        self.assertEqual(created_observation.get('message'), "Observations uploaded")
        self.assertEqual(created_observation.get('status'), 201)

        self.api.destroy_observations(streamid=s.id)
        self.api.destroy_stream(id=s.id)

    def generate_observations(self):

        o = Observation()

        dt = datetime.datetime(2016, 2, 15, 0, 0, 0)
        dt_td = datetime.timedelta(minutes=15)

        points = [
            {'time': dt + (dt_td * 0), 'v': 1},
            {'time': dt + (dt_td * 1), 'v': 2},
            {'time': dt + (dt_td * 2), 'v': 3},
        ]

        for p in points:
            item = UnivariateResult()
            item.t = p.get('time')
            item.v = {
                'v': p.get('v')
            }
            o.results.append(item)

        return o, points

    def generate_document_json_observations(self):

        o = Observation()

        dt = datetime.datetime(2016, 2, 15, 0, 0, 0)
        dt_td = datetime.timedelta(minutes=15)

        points = [
            {'time': dt + (dt_td * 0), 'v': {'j': {'propa': 'valuea', 'propb': 1}}},
            {'time': dt + (dt_td * 1), 'v': {'j': {'propa': 'valueb', 'propb': 2}}},
            {'time': dt + (dt_td * 2), 'v': {'j': {'propa': 'valuec', 'propb': 3}}},
        ]

        for p in points:
            item = UnivariateResult()
            item.t = p.get('time')
            item.v = p.get('v')
            o.results.append(item)

        return o, points

    def generate_document_text_observations(self):

        o = Observation()

        dt = datetime.datetime(2016, 2, 15, 0, 0, 0)
        dt_td = datetime.timedelta(minutes=15)

        points = [
            {'time': dt + (dt_td * 0), 'v': {'d': "blah"}},
            {'time': dt + (dt_td * 1), 'v': {'d': "blah2"}},
            {'time': dt + (dt_td * 2), 'v': {'d': "blah3\nblah4"}},
        ]

        for p in points:
            item = UnivariateResult()
            item.t = p.get('time')
            item.v = p.get('v')
            o.results.append(item)

        return o, points

    def generate_observations_with_annoations(self):

        o = Observation()

        dt = datetime.datetime(2016, 2, 15, 0, 0, 0)
        dt_td = datetime.timedelta(minutes=15)

        points = [
            {'time': dt + (dt_td * 0), 'v': 1},
            {'time': dt + (dt_td * 1), 'v': 2},
            {'time': dt + (dt_td * 2), 'v': 3},
        ]

        for p in points:
            item = UnivariateResult()
            item.t = p.get('time')
            item.v = {
                'v': p.get('v'),
                'a': {
                    'key1': 'value1',
                    'key2': 'value2'
                }
            }
            o.results.append(item)

        return o, points

    @tape.use_cassette('test_create_scalar_observations.json')
    def test_create_scalar_observations(self):
        s = self.generate_scalar_stream()

        o, points = self.generate_observations()

        dt_format = '%Y-%m-%dT%H:%M:%S.%fZ'
        required_state = {
            'results': [
                {
                    't': points[0].get('time').strftime(dt_format),
                    'v': {
                        'v': points[0].get('v')
                    }
                },
                {
                    't': points[1].get('time').strftime(dt_format),
                    'v': {
                        'v': points[1].get('v')
                    }
                },
                {
                    't': points[2].get('time').strftime(dt_format),
                    'v': {
                        'v': points[2].get('v')
                    }
                },
            ]
        }
        required_json = dumps(required_state, sort_keys=True)  # be explict with key order since dumps gives us a string

        actual_state = o.to_state("create")
        actual_json = o.to_json("create")

        ### dict diff debugging
        # from deepdiff import DeepDiff
        # diff = DeepDiff(required_state, actual_state)

        # verify json
        self.assertEqual(actual_json, required_json)

        print("creating stream %s" % s)
        created_stream = self.api.create_stream(s)

        self.assertEqual(created_stream.id, s.id)

        created_observation = self.api.create_observations(o, streamid=s.id)

        self.assertEqual(created_observation.get('message'), "Observations uploaded")
        self.assertEqual(created_observation.get('status'), 201)

        self.api.destroy_observations(streamid=s.id)
        self.api.destroy_stream(id=s.id)

    @tape.use_cassette('test_create_scalar_observations_with_annotations.json')
    def test_create_scalar_observations_with_annotations(self):
        s = self.generate_scalar_stream()

        o, points = self.generate_observations_with_annoations()

        dt_format = '%Y-%m-%dT%H:%M:%S.%fZ'
        required_state = {
            'results': [
                {
                    't': points[0].get('time').strftime(dt_format),
                    'v': {
                        'v': points[0].get('v'),
                        'a': {
                            'key1': 'value1',
                            'key2': 'value2'
                        }
                    }
                },
                {
                    't': points[1].get('time').strftime(dt_format),
                    'v': {
                        'v': points[1].get('v'),
                        'a': {
                            'key1': 'value1',
                            'key2': 'value2'
                        }
                    }
                },
                {
                    't': points[2].get('time').strftime(dt_format),
                    'v': {
                        'v': points[2].get('v'),
                        'a': {
                            'key1': 'value1',
                            'key2': 'value2'
                        }
                    }
                },
            ]
        }
        required_json = dumps(required_state, sort_keys=True)  # be explict with key order since dumps gives us a string

        actual_state = o.to_state("create")
        actual_json = o.to_json("create")

        ### dict diff debugging
        # from deepdiff import DeepDiff
        # diff = DeepDiff(required_state, actual_state)

        # verify json
        self.assertEqual(actual_json, required_json)

        print("creating stream %s" % s)
        created_stream = self.api.create_stream(s)

        self.assertEqual(created_stream.id, s.id)

        created_observation = self.api.create_observations(o, streamid=s.id)

        self.assertEqual(created_observation.get('message'), "Observations uploaded")
        self.assertEqual(created_observation.get('status'), 201)

        self.api.destroy_observations(streamid=s.id)
        self.api.destroy_stream(id=s.id)

    @tape.use_cassette('test_create_vector_observations.json')
    def test_create_vector_observations(self):
        o = Observation()

        dt = datetime.datetime(2016, 2, 15, 0, 0, 0)
        dt_td = datetime.timedelta(minutes=15)

        points = [1, 2, 3]

        s = self.generate_vector_stream(len(points))

        item = UnivariateResult()
        item.t = dt
        item.v = {'v': points}
        o.results.append(item)

        dt_format = '%Y-%m-%dT%H:%M:%S.%fZ'
        required_state = {
            'results': [
                {
                    't': dt.strftime(dt_format),
                    'v': {'v': [points[0], points[1], points[2]]}
                },
            ]
        }
        required_json = dumps(required_state,
                              sort_keys=True)  # be explict with key order since dumps gives us a string

        actual_state = o.to_state("create")
        actual_json = o.to_json("create")

        ### dict diff debugging
        # from deepdiff import DeepDiff
        # diff = DeepDiff(required_state, actual_state)

        # verify json
        self.assertEqual(actual_json, required_json)

        print("creating stream %s" % s)
        created_stream = self.api.create_stream(s)

        self.assertEqual(created_stream.id, s.id)

        created_observation = self.api.create_observations(o, streamid=s.id)

        self.assertEqual(created_observation.get('message'), "Observations uploaded")
        self.assertEqual(created_observation.get('status'), 201)

        self.api.destroy_observations(streamid=s.id)
        self.api.destroy_stream(id=s.id)

    @tape.use_cassette('test_create_regularly_binned_vector_observations.json')
    def test_create_regularly_binned_vector_observations(self):
        o = Observation()

        dt = datetime.datetime(2016, 2, 15, 0, 0, 0)
        dt_td = datetime.timedelta(minutes=15)

        points = [1, 2, 3]

        s = self.generate_regularly_binned_vector_stream(1, 3, 1, "a8a8ce25-30f6-4b1a-ac78-533d2887280f")

        item = UnivariateResult()
        item.t = dt
        item.v = {'v': points}
        o.results.append(item)

        dt_format = '%Y-%m-%dT%H:%M:%S.%fZ'
        required_state = {
            'results': [
                {
                    't': dt.strftime(dt_format),
                    'v': {'v': [points[0], points[1], points[2]]}
                },
            ]
        }
        required_json = dumps(required_state,
                              sort_keys=True)  # be explict with key order since dumps gives us a string

        actual_state = o.to_state("create")
        actual_json = o.to_json("create")

        ### dict diff debugging
        # from deepdiff import DeepDiff
        # diff = DeepDiff(required_state, actual_state)

        # verify json
        self.assertEqual(actual_json, required_json)

        print("creating stream %s" % s)
        created_stream = self.api.create_stream(s)

        self.assertEqual(created_stream.id, s.id)

        created_observation = self.api.create_observations(o, streamid=s.id)

        self.assertEqual(created_observation.get('message'), "Observations uploaded")
        self.assertEqual(created_observation.get('status'), 201)

        self.api.destroy_observations(streamid=s.id)
        self.api.destroy_stream(id=s.id)

    @tape.use_cassette('test_create_image_observations.json')
    def test_create_image_observations(self):
        s = self.generate_image_stream("13136661-8c66-47c6-9cd1-b74e4214a4ab")

        o = Observation()

        dt = datetime.datetime(2016, 2, 15, 0, 0, 0)
        dt_td = datetime.timedelta(minutes=15)

        points = [
            {'time': dt + (dt_td * 0), 'm': 'image/png',
             'd': 'iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAYAAAAfFcSJAAAADUlEQVR42mNk+M9QDwADhgGAWjR9awAAAABJRU5ErkJggg=='},
            {'time': dt + (dt_td * 1), 'm': 'image/png',
             'd': 'iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAYAAAAfFcSJAAAADUlEQVR42mNk+M9QDwADhgGAWjR9awAAAABJRU5ErkJggg=='},
            {'time': dt + (dt_td * 2), 'm': 'image/png',
             'd': 'iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAYAAAAfFcSJAAAADUlEQVR42mNk+M9QDwADhgGAWjR9awAAAABJRU5ErkJggg=='}
        ]

        for p in points:
            item = UnivariateResult()
            item.t = p.get('time')
            item.v = {
                'm': p.get('m'),
                'd': p.get('d')
            }
            o.results.append(item)

        dt_format = '%Y-%m-%dT%H:%M:%S.%fZ'
        required_state = {
            'results': [
                {
                    't': points[0].get('time').strftime(dt_format),
                    'v': {
                        'm': points[0].get('m'),
                        'd': points[0].get('d')
                    }
                },
                {
                    't': points[1].get('time').strftime(dt_format),
                    'v': {
                        'm': points[1].get('m'),
                        'd': points[1].get('d')
                    }
                },
                {
                    't': points[2].get('time').strftime(dt_format),
                    'v': {
                        'm': points[2].get('m'),
                        'd': points[2].get('d')
                    }
                },
            ]
        }
        required_json = dumps(required_state, sort_keys=True)  # be explict with key order since dumps gives us a string

        actual_state = o.to_state("create")
        actual_json = o.to_json("create")

        ### dict diff debugging
        # from deepdiff import DeepDiff
        # diff = DeepDiff(required_state, actual_state)

        # verify json
        self.assertEqual(actual_json, required_json)

        print("creating stream %s" % s.id)
        created_stream = self.api.create_stream(s)
        print("done creating stream %s")

        self.assertEqual(created_stream.id, s.id)

        print("creating observations %s" % s.id)
        created_observation = self.api.create_observations(o, streamid=s.id)

        print("done creating observations %s" % s.id)
        self.assertEqual(created_observation.get('message'), "Observations uploaded")
        self.assertEqual(created_observation.get('status'), 201)

        self.api.destroy_observations(streamid=s.id)
        self.api.destroy_stream(id=s.id)

    @tape.use_cassette('test_create_document_json_observations.json')
    def test_create_document_json_observations(self):
        s = self.generate_document_stream()

        o, points = self.generate_document_json_observations()

        dt_format = '%Y-%m-%dT%H:%M:%S.%fZ'
        required_state = {
            'results': [
                {
                    't': points[0].get('time').strftime(dt_format),
                    'v': points[0].get('v')
                },
                {
                    't': points[1].get('time').strftime(dt_format),
                    'v': points[1].get('v')
                },
                {
                    't': points[2].get('time').strftime(dt_format),
                    'v': points[2].get('v')
                },
            ]
        }
        required_json = dumps(required_state, sort_keys=True)  # be explict with key order since dumps gives us a string

        actual_state = o.to_state("create")
        actual_json = o.to_json("create")

        ### dict diff debugging
        # from deepdiff import DeepDiff
        # diff = DeepDiff(required_state, actual_state)

        # verify json
        self.assertEqual(actual_json, required_json)

        print("creating stream %s" % s)
        created_stream = self.api.create_stream(s)

        self.assertEqual(created_stream.id, s.id)

        created_observation = self.api.create_observations(o, streamid=s.id)

        self.assertEqual(created_observation.get('message'), "Observations uploaded")
        self.assertEqual(created_observation.get('status'), 201)

        self.api.destroy_observations(streamid=s.id)
        self.api.destroy_stream(id=s.id)

    @tape.use_cassette('test_create_document_text_observations.json')
    def test_create_document_text_observations(self):
        s = self.generate_document_stream(mimetype=StreamMetaDataMimeType.text)

        o, points = self.generate_document_text_observations()

        dt_format = '%Y-%m-%dT%H:%M:%S.%fZ'
        required_state = {
            'results': [
                {
                    't': points[0].get('time').strftime(dt_format),
                    'v': points[0].get('v')
                },
                {
                    't': points[1].get('time').strftime(dt_format),
                    'v': points[1].get('v')
                },
                {
                    't': points[2].get('time').strftime(dt_format),
                    'v': points[2].get('v')
                },
            ]
        }
        required_json = dumps(required_state, sort_keys=True)  # be explict with key order since dumps gives us a string

        actual_state = o.to_state("create")
        actual_json = o.to_json("create")

        ### dict diff debugging
        # from deepdiff import DeepDiff
        # diff = DeepDiff(required_state, actual_state)

        # verify json
        self.assertEqual(actual_json, required_json)

        print("creating stream %s" % s)
        created_stream = self.api.create_stream(s)

        self.assertEqual(created_stream.id, s.id)

        created_observation = self.api.create_observations(o, streamid=s.id)

        self.assertEqual(created_observation.get('message'), "Observations uploaded")
        self.assertEqual(created_observation.get('status'), 201)

        self.api.destroy_observations(streamid=s.id)
        self.api.destroy_stream(id=s.id)

    @tape.use_cassette('test_connection_timeout.json')
    def test_connection_timeout(self):

        # Override connect timeout to speed this test up a bit
        connect_timeout = 1
        self.api.timeout = (connect_timeout, 3)
        self.api.backoff_factor = 0

        # Set to an invalid IP to force a connect timeout
        self.api.host = '1.2.3.4'

        # Calculate expected time to fail
        expected_time = connect_timeout * (self.api.connect_retries + 1)

        t0 = time.time()

        with self.assertRaises(SenapsError):
            self.api.streams()

        t = time.time() - t0
        print('Connect timeout took', t, 'seconds')
        self.assertTrue((expected_time - 0.5) <= t <= (expected_time + 0.5),
                        'Timeout took %f, expected, %f' % (t, expected_time))

    @tape.use_cassette('test_connection_refused.json')
    def test_connection_refused(self):

        # Set to localhost and closed port to force a connect refused
        self.api.host = 'localhost:700'  # hopefully a closed port

        # slow down the backoff, set connect retries
        self.api.backoff_factor = 2
        self.api.connect_retries = 3

        # Calculate expected time to fail
        expected_time = 12  # 0s, 4s, 8s = 12s total

        t0 = time.time()

        with self.assertRaises(SenapsError):
            self.api.streams()

        t = time.time() - t0
        print('Connect timeout took', t, 'seconds')
        self.assertTrue((expected_time - 0.5) <= t <= (expected_time + 0.5),
                        'Timeout took %f, expected, %f' % (t, expected_time))

    @tape.use_cassette('test_bad_gateway.json')
    def test_bad_gateway(self):

        # Set to localhost and closed port to force a connect refused
        self.api.host = 'httpstat.us'  # hopefully a closed port
        self.api.api_root = '/'

        # slow down the backoff, set connect retries
        self.api.backoff_factor = 2
        self.api.status_retries = 3

        # Calculate expected time to fail
        expected_time = 16
		# Tolerance was 1 but there is now a lot more variation
        expected_time_tolerance = 5

        t0 = time.time()

        with self.assertRaises(SenapsError):
            self.r = bind_api(
                api=self.api,
                path='/502',  # point us to the httpstat.us test endpoint
                payload_type='json',
                allowed_param=['limit', 'id'],
                query_only_param=[
                    'id',
                    'limit',
                    'skip',
                    'near',
                    'radius',
                    'resulttype',
                    'expand',
                    'recursive',
                    'groupids',
                    'organisationid',
                    'locationid',
                    'usermetadatafield',
                    'usermetadatavalues'
                ],
                payload_list=True,
                require_auth=True,
            )()

        t = time.time() - t0
        print('Connect timeout took', t, 'seconds')
        self.assertTrue((expected_time - expected_time_tolerance) <= t <= (expected_time + expected_time_tolerance),
                        'Timeout took %f, expected, %f' % (t, expected_time))

    def test_get_locations(self):

        locations = self.api.locations()

        self.assertGreater(len(locations.ids()), 100)
        self.assertIsNotNone(locations.ids()[0])

    def test_get_limited_locations(self):
        locations = self.api.locations(limit=1)

        self.assertEqual(len(locations.ids()), 1)

    def test_get_expanded_locations(self):
        locations = self.api.locations(id='integration_test', expand=True)

        self.assertEqual(len(locations.ids()), 1)
        self.assertEqual(locations[0].groups[0].id, 'integration_test')
        self.assertEqual(len(locations[0].organisations), 1)
        self.assertIsNotNone(locations.ids()[0])

    def test_get_location_by_id(self):
        test_loc_id = 'integration_test'
        location = self.api.get_location(id=test_loc_id)

        self.assertEqual(test_loc_id, location.id)
        self.assertEqual(location.groups[0].id, 'integration_test')
        self.assertEqual(len(location.organisations), 1)

    @unittest.skip('CPS-1027: expanded location queries will timeout right now.')
    def test_get_locations_expanded_includes_coordinates(self):

        locations = self.api.locations(expand=True, limit=10)

        self.assertGreater(len(locations.ids()), 100)
        self.assertIsNotNone(locations.ids()[0])
        self.assertEquals(len(locations[0].geojson['coordinates']), 2)

    @tape.use_cassette('test_search_with_no_results_returns_empty_list.json')
    def test_search_with_no_results_returns_empty_list(self):
        groups = self.api.get_groups(groupids='a_nonexistent_group')

        self.assertEquals(len(groups), 0)

    @tape.use_cassette('test_get_permitted.json')
    def test_get_permitted(self):
        """
        Params:
        "permission"
        "resourceid"
        "organisationid"
        "groupids"
        ----
        Request must provide:
        permission and (resourceid or (organisationid && groupids))
        """
        # tests rely on exported vars, set some inputs based on those.
        if host == 'senaps.io':
            # no sandbox org in production server.
            org_id = 'csiro'
            groupids = 'sandbox'
        else:
            org_id = 'sandbox'
            groupids = 'sandbox'
        permitted = self.api.get_permitted(permission='.ReadStreamPermission',
                                           organisationid=org_id,
                                           groupids=groupids)

        permitted_on_resource = self.api.get_permitted(permission='.ReadStreamPermission',
                                                       resourceid='empty_scalar_stream')
        self.assertTrue(permitted is not None)
        self.assertTrue(hasattr(permitted, 'permitted'))

        self.assertTrue(permitted_on_resource is not None)
        self.assertTrue(hasattr(permitted, 'permitted'))

    def test_get_all_roles(self):
        """
        Get a list of all roles defined on the server.
        Will not succeed unless you run these tests with an Administrative user.
        """
        roles = self.api.roles()
        self.assertGreater(len(roles), 1, 'expected more than 1 role to be found.')

    def test_get_expanded_roles(self):
        """
        Get a list of all roles defined on the server.
        Will not succeed unless you run these tests with an Administrative user.
        """
        roles = self.api.roles(expand=True)
        self.assertTrue(hasattr(roles[0], 'implicit'),
                        'Expand arg should cause an implicit member to be generated on Role objects')

    def test_get_role_by_id(self):
        """
        Check the 'admin' role exists.
        Will not succeed unless you run this with an Administrative user.
        """
        role = self.api.get_role('admin')
        self.assertTrue(role.id == 'admin')

    def test_create_and_delete_role(self):
        """
        Add a new role with a simple permission.
        Will not succeed unless user has '.CreateRolePermission' or '.UpdateRolePermission'
        """
        test_role_id = 'a_test_role'
        rg_perm = '.ReadGroupPermission'
        permissions = [{'type': rg_perm}]
        role_type = ".GroupRole"
        organisationid = "sandbox"
        groupid = "sandbox_group"
        addressfilters = []

        self.api.create_role(id='a_test_role',
                             permissions=permissions,
                             type=role_type,
                             groupid=groupid,
                             organisationid=organisationid,
                             addressfilters=addressfilters)
        the_role = self.api.get_role(test_role_id)
        self.api.delete_role(id=test_role_id)
        self.assertTrue(test_role_id == the_role.id, "expected a role with the correct ID to be retrieved after PUT")

    def test_create_and_update_role(self):
        """
        Add a new role with a simple permission.
        Will not succeed unless user has '.CreateRolePermission' or '.UpdateRolePermission'
        """
        test_role_id = 'a_test_role'
        rg_perm = '.ReadGroupPermission'
        permissions = [{'type': rg_perm}]
        role_type = ".GroupRole"
        organisationid = "sandbox"
        groupid = "sandbox_group"
        addressfilters = []

        self.api.create_role(id=test_role_id,
                             permissions=permissions,
                             type=role_type,
                             groupid=groupid,
                             organisationid=organisationid,
                             addressfilters=addressfilters)
        the_role = self.api.get_role(test_role_id)
        # role exists right now, let's update it.
        self.api.update_role(id=the_role.id,
                             permissions=the_role.permissions,
                             type=the_role.type,
                             groupid=the_role.groupid,
                             organisationid=the_role.organisationid,
                             addressfilters=['1.1.1.1'])  # patch addressfilters to see if update works.
        updated_role = self.api.get_role(test_role_id)
        self.assertTrue('1.1.1.1' in updated_role.addressfilters, 'Expected 1.1.1.1 to be added to address filters')
        self.api.delete_role(id=test_role_id)

    def test_delete_role(self):
        """
        We should be able to create and delete a role and then verify it is gone.
        Will not succeed unless user has '.CreateRolePermission' and '.DeleteRolePermission'
        """
        test_role_id = 'a_test_role_for_deletion'
        self.given_the_group_role(test_role_id, 'sandbox_group', 'sandbox', ['.ReadGroupPermission'], None)
        self.api.get_role(test_role_id)
        # if the above succeeded, we had permission to create and read back the role, validating rest of test.
        self.api.delete_role(id=test_role_id)
        # we expect it to go boom with a 403 here (indistinguishable between not found and no perms).
        self.assertRaises(SenapsError, self.api.get_role, test_role_id)

    def test_create_user_no_roles(self):
        """
        Attempt to PUT a new user in Senaps. Note: this does NOT create the authentication front-end to a user.
        User management routines are present for internal use only,
        as they require Administrative permissions not available
        to organisation or group roles.
        """
        userid = 'a_probably_valid.name@emailhost.fake'
        result = self.given_the_user(userid)
        self.api.delete_user(id=userid)
        self.assertEqual(result.id, userid, 'Expected to retrieve the same user back from Senaps.')

    def test_create_hidden_user(self):
        """
        PUT a new user into Senaps, marking them as hidden.
        Note: this does NOT create the authentication front-end to a user.
        User management routines are present for internal use only,
        as they require Administrative permissions not available
        to organisation or group roles.
        """
        userid = 'a_hidden_user@emailhost.fake'
        result = self.given_the_user(userid, hidden=True)
        self.api.delete_user(id=userid)
        self.assertTrue(result.hidden, 'Expected %s to be a hidden user' % userid)

    @unittest.skip('Production server does not support this at the time of creating this test.')
    def test_create_user_with_eula(self):
        """
        PUT a new user into Senaps, accepting a eula.
        Note: this does NOT create the authentication front-end to a user.
        User management routines are present for internal use only,
        as they require Administrative permissions not available
        to organisation or group roles.
        """
        userid = 'a_user_with_eula_user@emailhost.fake'
        result = self.given_the_user(userid, eulaids=['senaps-eula-v1'])
        self.api.delete_user(id=userid)
        self.assertEqual('senaps-eula-v1', result._json.get('eulaids')[0],
                         'Expected the user to have the senaps eula on the list of eulas')

    def test_create_user_no_hidden_argument(self):
        """
        PUT a new user into Senaps, but do not specify whether they should be hidden or not.
        Note: this does NOT create the authentication front-end to a user.
        User management routines are present for internal use only,
        as they require Administrative permissions not available
        to organisation or group roles.
        """
        userid = 'a_partially_defined_user@emailhost.fake'
        result = self.api.create_user(id=userid)
        self.api.delete_user(id=userid)
        self.assertFalse(result.hidden, 'Expected %s to not be hidden.')

    def test_create_user_simple_group_role(self):
        """
        Attempt to PUT a new user in Senaps. Note: this does NOT create the authentication front-end to a user.
        User management routines are present for internal use only,
         as they require Administrative permissions not available
        to organisation or group roles.
        """
        userid = 'definitely_a_test_throwaway.name@emailhost.fake'
        role_id = 'a_nice_deleteable_role_id'
        self.given_the_group_role(role_id,
                                  'sandbox_group',
                                  'sandbox',
                                  ['.ReadGroupPermission'], None)

        result = self.given_the_user(userid, roles=[role_id])
        self.api.delete_user(id=userid)
        self.api.delete_role(id=role_id)
        self.assertEqual(role_id, result.roles[0].id, 'Expected the user to have the "%s" role' % role_id)
        self.assertEqual(result.id, userid, 'Expected to retrieve the same user back from Senaps.')

    def test_delete_user(self):
        """
        Deletion of a user should result in a resource no longer being found.
        User management routines are present for internal use only,
         as they require Administrative permissions not available
        to organisation or group roles.
        """
        # while we delete users during other tests, we don't actually check to make sure they went away...do so here.
        userid = 'will_be_deleted@emailhost.fake'
        self.given_the_user(userid)
        self.api.delete_user(id=userid)
        self.assertRaises(SenapsError, self.api.get_user, userid)

    def test_update_user(self):
        """
        Create and then change a user's assigned roles.
        User management routines are present for internal use only,
        as they require Administrative permissions not available
        to organisation or group roles.
        """
        userid = 'updateable_user@emailhost.fake'
        throwaway_role_id = 'a_throwaway_group_role_id'
        self.given_the_user(userid)
        self.given_the_group_role(throwaway_role_id,
                                  'sandbox_group',
                                  'sandbox',
                                  ['.ReadGroupPermission'], None)
        self.api.update_user(id=userid, roleids=[throwaway_role_id])
        # verify it was updated.
        retrieved = self.api.get_user(userid)
        self.api.delete_user(id=userid)
        self.assertTrue(retrieved.roles[0].id == throwaway_role_id,
                        'Expected %s as a role on %s user' % (throwaway_role_id, userid))

    @unittest.skip('Production server does not support this at the time of creating this test.')
    def test_update_user_with_eula(self):
        """
        Create and then change a user's accepted eulas.
        User management routines are present for internal use only,
        as they require Administrative permissions not available
        to organisation or group roles.
        """
        userid = 'updateable_user@emailhost.fake'
        throwaway_role_id = 'a_throwaway_group_role_id'
        self.given_the_user(userid)
        self.api.update_user(id=userid, eulaids=['senaps-eula-v1'])
        # verify it was updated.
        retrieved = self.api.get_user(userid)
        self.api.delete_user(id=userid)
        self.assertEqual(1, len(retrieved._json.get('eulaids')), 'Expected the total number of accepted eulas to be 1.')
        self.assertEqual('senaps-eula-v1', retrieved._json.get('eulaids')[0],
                         'Expected senaps-eula-v1 as accepted eula after update')

    def test_update_user_hidden(self):
        """
        Create and then update a user, marking them as hidden. Hidden users are used for machine accounts in Senaps.
        User management routines are present for internal use only,
        as they require Administrative permissions not available
        to organisation or group roles.
        """
        userid = 'updateable_user@emailhost.fake'
        throwaway_role_id = 'a_throwaway_group_role_id'
        new_user = self.given_the_user(userid)  # helper routine always makes users with hidden = False.
        self.given_the_group_role(throwaway_role_id,
                                  'sandbox_group',
                                  'sandbox',
                                  ['.ReadGroupPermission'], None)
        self.api.update_user(id=userid, hidden=True, roleids=[throwaway_role_id])
        # verify it was updated.
        retrieved = self.api.get_user(userid)
        self.api.delete_user(id=userid)
        self.assertTrue(retrieved.hidden,
                        'Expected %s user to be hidden' % throwaway_role_id)

    def test_get_all_users(self):
        """
        Retrieve a list of all users in the system.
        """
        # no filtering.
        users = self.api.users()
        self.assertTrue(len(users) > 0,
                        'Expected at least 1 user in your test Senaps instance (e.g. Invoker of this call)')

    def test_get_user_by_filtering(self):
        """
        Retrieve a user by filtering by ID.
        """
        expected_user = 'add_a_little_chaos@emailhost.fake'
        self.given_the_user(expected_user)
        users = self.api.users(id=expected_user)
        self.api.delete_user(id=expected_user)
        self.assertTrue(len(users) == 1,
                        'Expected at least 1 user in your test Senaps instance (e.g. Invoker of this call)')

    def test_get_user_by_filtering_on_roles(self):
        """
        Retrieve a user by filtering by roleids.
        """
        expected_user = 'add_a_little_chaos@emailhost.fake'
        throwaway_role_id = 'throwaway_role_id'
        self.given_the_group_role(throwaway_role_id, 'sandbox_group', 'sandbox', ['.ReadGroupPermission'])
        self.given_the_user(expected_user, roles=[throwaway_role_id])

        users = self.api.users(roleids=[throwaway_role_id])
        self.api.delete_user(id=expected_user)
        self.api.delete_role(id=throwaway_role_id)
        self.assertTrue(len(users) == 1,
                        'Expected at least 1 user in your test Senaps instance (e.g. Invoker of this call)')

    def test_round_tripping_group_user_metadata(self):
        user_metadata = {'test': 'value'}

        self.api.create_group(id='user_metadata_group', name='User metadata test group', organisationid='sandbox',
                              usermetadata=user_metadata)

        group = self.api.get_group('user_metadata_group')
        self.api.destroy_group('user_metadata_group')

        self.assertEqual(user_metadata, group['usermetadata'])

    def given_the_user(self, userid, hidden=False, roles=None, eulaids=None):
        """
        Invoke the User PUT verb on your chosen test server.
        :param userid: str: a valid userid in Senaps
        :param roles: list: a list of Role IDs that should exist in Senaps.
        :return: User object, or raises a SenapsError if invalid permissions.
        """
        arguments = dict(id=userid,
                         hidden=hidden)
        if roles is not None:
            arguments['roleids'] = roles
        if eulaids is not None:
            arguments['eulaids'] = eulaids
        return self.api.create_user(**arguments)

    def given_the_group_role(self, roleid, groupid, organisationid, permissions, addressfilters=None):
        """
        Invoke the Role PUT verb on your chosen test server
        :param roleid: str: roleid to create.
        :param groupid: str: a valid group id in Senaps
        :param organisationid: str: a valid organisation in Senaps
        :param permissions: list: a list of Senaps permissions
        :param addressfilters: list: defaults to None, a list of IPv4 addresses that are permitted to access this role.
        :return: Role object, or raises a Senaps error if not permitted to undertake this function.
        """

        perms = [{'type': p} for p in permissions]

        if addressfilters is None:
            return self.api.create_role(id=roleid,
                                        groupid=groupid,
                                        organisationid=organisationid,
                                        type='.GroupRole',
                                        permissions=perms)

        return self.api.create_role(id=roleid,
                                    groupid=groupid,
                                    organisationid=organisationid,
                                    type='.GroupRole',
                                    permissions=perms,
                                    addressfilters=addressfilters)


class TestAPIConnectionProtocol(unittest.TestCase):

    def test_valid_protocols(self):
        for protocol in VALID_PROTOCOLS:
            auth = HTTPConsumerIDAuth('myfakeuser@domainname.com')
            api = API(auth, protocol=protocol)

    def test_invalid_protocol_will_raise(self):
        auth = HTTPConsumerIDAuth('myfakeuser@domainname.com')
        with self.assertRaises(ValueError):
            API(auth, protocol='git')



"""
MIT License
Copyright (c) 2016 Ionata Digital
Copyright (c) 2009-2014 Joshua Roesslein

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from __future__ import print_function

import time
import re
from collections import OrderedDict

import six
import requests
import logging

from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

from senaps_sensor.error import SenapsError, RateLimitError, is_rate_limit_error_message
from senaps_sensor.utils import convert_to_utf8_str
from senaps_sensor.models import Model

if six.PY2:
    from urllib import quote
else:
    from urllib.parse import quote

re_path_template = re.compile('{\w+}')

log = logging.getLogger('senset.binder')

try:
    from sensetdp.models import Model as SenseTModel
    
    model_classes = (Model, SenseTModel)
except ImportError:
    model_classes = (Model,)


def bind_api(**config):
    class APIMethod(object):

        api = config['api']
        path = config['path']
        action = config.get('action', None)
        payload_type = config.get('payload_type', None)
        payload_list = config.get('payload_list', False)
        allowed_param = config.get('allowed_param', [])
        query_only_param = config.get('query_only_param', [])
        method = config.get('method', 'GET')
        require_auth = config.get('require_auth', False)
        use_cache = config.get('use_cache', True)
        session = requests.Session()

        def __init__(self, args, kwargs):
            api = self.api
            self.api_root = api.api_root
            self.session.verify = api.verify

            # If authentication is required and no credentials
            # are provided, throw an error.
            if self.require_auth and not api.auth:
                raise SenapsError('Authentication required!')

            self.post_data = kwargs.pop('post_data', None)
            self.json_data = kwargs.pop('json_data', {})
            self.use_json = kwargs.pop('use_json', True)
            self.query_params = kwargs.pop('query_params', {})

            self.retry_count = kwargs.pop('retry_count',
                                          api.retry_count)
            self.retry_delay = kwargs.pop('retry_delay',
                                          api.retry_delay)
            self.retry_errors = kwargs.pop('retry_errors',
                                           api.retry_errors)
            self.wait_on_rate_limit = kwargs.pop('wait_on_rate_limit',
                                                 api.wait_on_rate_limit)
            self.wait_on_rate_limit_notify = kwargs.pop('wait_on_rate_limit_notify',
                                                        api.wait_on_rate_limit_notify)
            self.parser = kwargs.pop('parser', api.parser)
            self.session.headers = kwargs.pop('headers', {})

            self.session = self.requests_retry_session(kwargs.pop('connect_retries', api.connect_retries),
                                                       kwargs.pop('read_retries', api.read_retries),
                                                       kwargs.pop('backoff_factor', api.backoff_factor),
                                                       kwargs.pop('status_retries', api.status_retries),
                                                       session=self.session)

            self.build_data(args, kwargs)
            self.build_query_params(kwargs)

            # Perform any path variable substitution
            self.build_path()

            self.host = api.host
            self.host_protocol = api.protocol

            # TODO: test and remove below.
            # Manually set Host header to fix an issue in python 2.5
            # or older where Host is set including the 443 port.
            # This causes Twitter to issue 301 redirect.
            # See Issue https://github.com/tweepy/tweepy/issues/12
            # self.session.headers['Host'] = self.host

            # Monitoring rate limits
            self._remaining_calls = None
            self._reset_time = None

        def requests_retry_session(self, connect_retries, read_retries, backoff_factor, status_retries,
                                   status_forcelist=(500, 502, 504), session=None):

            session = session or requests.Session()
            retry = Retry(
                total=None,
                read=read_retries,
                connect=connect_retries,
                status=status_retries,
                backoff_factor=backoff_factor,
                status_forcelist=status_forcelist,
            )

            adapter = HTTPAdapter(max_retries=retry)
            session.mount('http://', adapter)
            session.mount('https://', adapter)
            return session

        def build_data(self, args, kwargs):
            if len(args) == 1 and isinstance(args[0], model_classes):
                # explode model.to_state() of model instance into kwargs, clear args
                kwargs.update(args[0].to_state(self.action))
                args = list()
            else:
                for k, v in kwargs.items():
                    if isinstance(v, model_classes):
                        kwargs[k] = v.to_state(self.action)

            # filter kwargs for allowed_param and not in query_only_param
            kwargs = dict([(k, v) for k, v in kwargs.items() if k in self.allowed_param])

            if self.use_json:
                self.json_data = dict([(k, v) for k, v in kwargs.items() if k not in self.query_only_param])

            self.session.params = OrderedDict()
            for idx, arg in enumerate(args):
                if arg is None:
                    continue
                try:
                    self.session.params[self.allowed_param[idx]] = convert_to_utf8_str(arg)
                except IndexError:
                    raise SenapsError('Too many parameters supplied!')

            for k, arg in kwargs.items():
                if arg is None:
                    continue
                if k in self.session.params:
                    raise SenapsError('Multiple values for parameter %s supplied!' % k)
                self.session.params[k] = convert_to_utf8_str(arg)

        def build_query_params(self, kwargs):
            for param in self.query_only_param:
                try:
                    self.query_params[param] = kwargs.get(param)
                except KeyError:
                    raise SenapsError("A required API.bind() method query_param was missing from the kwargs.")

        def build_path(self):
            for variable in re_path_template.findall(self.path):
                name = variable.strip('{}')

                if name == 'user' and 'user' not in self.session.params and self.api.auth:
                    # No 'user' parameter provided, fetch it from Auth instead.
                    value = self.api.auth.get_username()
                else:
                    try:
                        value = quote(self.session.params[name])
                    except KeyError:
                        raise SenapsError('No parameter value found for path variable: %s' % name)
                    del self.session.params[name]

                self.path = self.path.replace(variable, value)

            log.debug("PATH: %r", self.path)

        def execute(self):
            self.api.cached_result = False

            # Build the request URL
            url = self.api_root + self.path
            full_url = ('%s://' % self.host_protocol) + self.host + url

            # Query the cache if one is available
            # and this request uses a GET method.
            if self.use_cache and self.api.cache and self.method == 'GET':
                cache_result = self.api.cache.get(url)
                # if cache result found and not expired, return it
                if cache_result:
                    # must restore api reference
                    if isinstance(cache_result, list):
                        for result in cache_result:
                            if isinstance(result, model_classes):
                                result._api = self.api
                    else:
                        if isinstance(cache_result, model_classes):
                            cache_result._api = self.api
                    self.api.cached_result = True
                    return cache_result

            # Continue attempting request until successful
            # or maximum number of retries is reached.
            retries_performed = 0
            while retries_performed < self.retry_count + 1:
                # handle running out of api calls
                if self.wait_on_rate_limit:
                    if self._reset_time is not None:
                        if self._remaining_calls is not None:
                            if self._remaining_calls < 1:
                                sleep_time = self._reset_time - int(time.time())
                                if sleep_time > 0:
                                    if self.wait_on_rate_limit_notify:
                                        print("Rate limit reached. Sleeping for:", sleep_time)
                                    time.sleep(sleep_time + 5)  # sleep for few extra sec

                # if self.wait_on_rate_limit and self._reset_time is not None and \
                #                 self._remaining_calls is not None and self._remaining_calls < 1:
                #     sleep_time = self._reset_time - int(time.time())
                #     if sleep_time > 0:
                #         if self.wait_on_rate_limit_notify:
                #             print("Rate limit reached. Sleeping for: " + str(sleep_time))
                #         time.sleep(sleep_time + 5)  # sleep for few extra sec

                # Request compression if configured
                if self.api.compression:
                    self.session.headers['Accept-encoding'] = 'gzip'

                # Execute request
                try:
                    if self.use_json:
                        self.session.params = OrderedDict()
                        resp = self.session.request(self.method,
                                                    full_url,
                                                    json=self.json_data,
                                                    params=self.query_params,
                                                    timeout=self.api.timeout,
                                                    auth=self.api.auth,
                                                    proxies=self.api.proxy)
                    else:
                        resp = self.session.request(self.method,
                                                    full_url,
                                                    data=self.post_data,
                                                    params=self.query_params,
                                                    timeout=self.api.timeout,
                                                    auth=self.api.auth,
                                                    proxies=self.api.proxy)
                except Exception as e:
                    raise SenapsError('Failed to send request: %s' % e)
                rem_calls = resp.headers.get('x-rate-limit-remaining')
                if rem_calls is not None:
                    self._remaining_calls = int(rem_calls)
                elif isinstance(self._remaining_calls, int):
                    self._remaining_calls -= 1
                reset_time = resp.headers.get('x-rate-limit-reset')
                if reset_time is not None:
                    self._reset_time = int(reset_time)
                if self.wait_on_rate_limit and self._remaining_calls == 0 and (
                        # if ran out of calls before waiting switching retry last call
                        resp.status_code == 429 or resp.status_code == 420):
                    continue
                retry_delay = self.retry_delay
                # Exit request loop if non-retry error code
                if resp.status_code == 200:
                    break
                elif (resp.status_code == 429 or resp.status_code == 420) and self.wait_on_rate_limit:
                    if 'retry-after' in resp.headers:
                        retry_delay = float(resp.headers['retry-after'])
                elif self.retry_errors and resp.status_code not in self.retry_errors:
                    break

                retries_performed += 1

                # Sleep before retrying request again
                if retries_performed < self.retry_count + 1:  # Only sleep when not on the last retry
                    time.sleep(retry_delay)

            # If an error was returned, throw an exception
            self.api.last_response = resp
            if resp.status_code and not 200 <= resp.status_code < 300:
                try:
                    error_msg, api_error_code = \
                        self.parser.parse_error(resp.text)
                except Exception as ex:
                    error_msg = "Senaps error response: status code = %s" % resp.status_code
                    api_error_code = None

                if is_rate_limit_error_message(error_msg):
                    raise RateLimitError(error_msg, resp)
                else:
                    raise SenapsError(error_msg, resp, api_code=api_error_code)

            # Parse the response payload
            result = self.parser.parse(self, resp.text)

            # Store result into cache if one is available.
            if self.use_cache and self.api.cache and self.method == 'GET' and result:
                self.api.cache.store(url, result)

            return result

    def _call(*args, **kwargs):
        method = APIMethod(args, kwargs)
        if kwargs.get('create'):
            return method
        else:
            return method.execute()

    # Set pagination mode
    if 'cursor' in APIMethod.allowed_param:
        _call.pagination_mode = 'cursor'
    elif 'max_id' in APIMethod.allowed_param:
        if 'since_id' in APIMethod.allowed_param:
            _call.pagination_mode = 'id'
    elif 'page' in APIMethod.allowed_param:
        _call.pagination_mode = 'page'

    return _call



"""
MIT License
Copyright (c) 2020 CSIRO

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""
from __future__ import unicode_literals, absolute_import, print_function

import json
import datetime
import uuid

from senaps_sensor.models import Organisation,Stream, StreamResultType, StreamMetaData, \
    StreamMetaDataType, \
    InterpolationType, Observation, UnivariateResult
import pandas as pd
from pandas.testing import assert_frame_equal
from senaps_sensor.parsers import PandasObservationParser


from senaps_sensor.utils import SenseTEncoder

from tests.config import *

import six

if six.PY3:
    import unittest
    from unittest.case import skip
    UTC_INFO = datetime.timezone.utc
    CHECK_INFERRED_TYPES = True
else:
    import unittest2 as unittest
    from unittest2.case import skip
    import pytz
    # no such thing as datetime.timezone.utc in python 2
    UTC_INFO = pytz.utc
    CHECK_INFERRED_TYPES = False


def dumps(*args, **kwargs):
    if 'cls' not in kwargs:
        kwargs['cls'] = SenseTEncoder
    return json.dumps(*args, **kwargs)


class ParsersTestCase(SensorApiTestCase):

    def setUp(self):
        super(ParsersTestCase, self).setUp()

    def generate_scalar_stream(self, stream_id=None):
        sm = StreamMetaData()
        sm.type = StreamMetaDataType.scalar
        sm.interpolation_type = InterpolationType.continuous
        sm.observed_property = 'http://registry.it.csiro.au/def/environment/property/air_temperature'
        sm.unit_of_measure = 'http://registry.it.csiro.au/def/qudt/1.1/qudt-unit/DegreeCelsius'
        s = self._generate_stream(StreamResultType.scalar, sm, stream_id)
        return s

    def generate_vector_stream(self, length, stream_id=None):
        sm = StreamMetaData()
        sm.type = StreamMetaDataType.vector
        sm.length = length
        s = self._generate_stream(StreamResultType.vector, sm, stream_id)
        return s

    def _generate_stream(self, stream_type, stream_meta_data, stream_id=None):

        o = Organisation()
        o.id = 'sandbox'

        s = Stream()
        s.id = stream_id if stream_id is not None else str(uuid.uuid1())

        s.organisations = [o]

        s.result_type = stream_type

        s.samplePeriod = 'PT10S'
        s.reportingPeriod = 'P1D'

        s.metadata = stream_meta_data

        return s

    def generate_observations(self, basetimestamp=None, deltat=None, obs=None):

        o = Observation()
        if basetimestamp:
            dt = basetimestamp.replace(tzinfo=UTC_INFO)
        else:
            dt = datetime.datetime(2016, 2, 15, 0, 0, 0, tzinfo=UTC_INFO)

        if deltat:
            dt_td = deltat
        else:
            dt_td = datetime.timedelta(minutes=15)

        if obs:
            points = [{'time': dt + (dt_td * index), 'v': val} for index, val in enumerate(obs)]
        else:
            points = [
                {'time': dt + (dt_td * 0), 'v': 1},
                {'time': dt + (dt_td * 1), 'v': 2},
                {'time': dt + (dt_td * 2), 'v': 3},
            ]

        for p in points:
            item = UnivariateResult()
            item.t = p.get('time')
            item.v = {
                'v': p.get('v')
            }
            o.results.append(item)

        df = pd.DataFrame(points)
        df.set_index('time', inplace=True)
        df.index.names = ['timestamp']
        if isinstance(points[0]['v'],list):
            df = pd.DataFrame(df['v'].to_list(), index = df.index, columns=['v[%d]'%i for i in range(len(df['v'][0]))])
        return o, points, df


    @tape.use_cassette('test_pandas_get_observations_single_scalar_stream.json')
    def test_pandas_get_observations_single_scalar_stream(self):
        s = self.generate_scalar_stream()
        o, points, df = self.generate_observations(obs=[1.0,2.0,3.0])

        mapnames = {'v': s.id}
        expected_df = df.rename(columns=mapnames)
        print('expected dataframe')
        print(expected_df)

        print('creating stream and observations %s' % s.id)
        self.api.create_stream(s)
        self.api.create_observations(o, streamid=s.id)

        print('getting observations')
        CSVparser = PandasObservationParser()
        retrieived_observations = self.api.get_observations(streamid=s.id, media='csv', parser=CSVparser)
        print(retrieived_observations)

        assert_frame_equal(expected_df, retrieived_observations)

        self.api.destroy_observations(streamid=s.id)
        self.api.destroy_stream(id=s.id)

    @tape.use_cassette('test_pandas_get_observations_two_scalar_stream.json')
    def test_pandas_get_observations_two_scalar_stream(self):
        s1 = self.generate_scalar_stream()
        o1, points1, df1 = self.generate_observations(basetimestamp = datetime.datetime(2016, 2, 15, 0, 0, 0, tzinfo=UTC_INFO),
                                                      deltat = datetime.timedelta(minutes=15),
                                                      obs=[1.0,2.0,3.0])
        mapnames = {'v': s1.id}
        df1 = df1.rename(columns=mapnames)

        s2 = self.generate_scalar_stream()
        o2, points2, df2 = self.generate_observations(basetimestamp = datetime.datetime(2016, 2, 15, 0, 15, 0, tzinfo=UTC_INFO),
                                                      deltat = datetime.timedelta(minutes=15),
                                                      obs=[1.5,2.5,3.5])
        mapnames = {'v': s2.id}
        df2 = df2.rename(columns=mapnames)

        expected_df1 = pd.merge(df1, df2, how='outer', left_index=True, right_index=True)
        expected_df2 = pd.merge(df2, df1, how='outer', left_index=True, right_index=True)
        print('expected dataframe1')
        print(expected_df1)
        print('expected dataframe2')
        print(expected_df2)

        print('creating stream and observations %s' % s1.id)
        self.api.create_stream(s1)
        self.api.create_observations(o1, streamid=s1.id)

        print('creating stream and observations %s' % s2.id)
        created_stream = self.api.create_stream(s2)
        created_observations = self.api.create_observations(o2, streamid=s2.id)

        CSVparser = PandasObservationParser()
        print('getting observations stream ordering #1')
        retrieived_observations1 = self.api.get_observations(streamid='%s,%s' % (s1.id,s2.id), media='csv', parser=CSVparser)
        print(retrieived_observations1)
        print('getting observations stream ordering #2')
        retrieived_observations2 = self.api.get_observations(streamid='%s,%s' % (s2.id,s1.id), media='csv', parser=CSVparser)
        print(retrieived_observations2)

        assert_frame_equal(expected_df1, retrieived_observations1)
        assert_frame_equal(expected_df2, retrieived_observations2)

        self.api.destroy_observations(streamid=s1.id)
        self.api.destroy_stream(id=s1.id)
        self.api.destroy_observations(streamid=s2.id)
        self.api.destroy_stream(id=s2.id)

    @tape.use_cassette('test_pandas_get_observations_single_vector_stream.json')
    def test_pandas_get_observations_single_vector_stream(self):
        s = self.generate_vector_stream(3)
        o, points, df = self.generate_observations(obs=[[1.0,2.0,3.0],[1.1,2.1,3.1],[1.2,2.2,3.2]])

        mapnames = {'v[%d]'%i : s.id+'[%d]'%i for i in range(3)}
        expected_df = df.rename(columns=mapnames)
        print('expected dataframe')
        print(expected_df)

        print('creating stream and observations %s' % s.id)
        self.api.create_stream(s)
        self.api.create_observations(o, streamid=s.id)

        print('getting observations')
        CSVparser = PandasObservationParser()
        retrieived_observations = self.api.get_observations(streamid=s.id, media='csv', parser=CSVparser)
        print(retrieived_observations)

        assert_frame_equal(expected_df, retrieived_observations)

        self.api.destroy_observations(streamid=s.id)
        self.api.destroy_stream(id=s.id)

    @tape.use_cassette('test_pandas_get_observations_two_vector_stream.json')
    def test_pandas_get_observations_two_vector_stream(self):
        s1 = self.generate_vector_stream(3)
        o1, points1, df1 = self.generate_observations(basetimestamp = datetime.datetime(2016, 2, 15, 0, 0, 0, tzinfo=UTC_INFO),
                                                      deltat = datetime.timedelta(minutes=15),
                                                      obs=[[1.0,2.0,3.0],[1.1,2.1,3.1],[1.2,2.2,3.2]])
        mapnames = {'v[%d]'%i : s1.id+'[%d]'%i for i in range(3)}
        df1 = df1.rename(columns=mapnames)

        s2 = self.generate_vector_stream(3)
        o2, points2, df2 = self.generate_observations(basetimestamp = datetime.datetime(2016, 2, 15, 0, 15, 0, tzinfo=UTC_INFO),
                                                      deltat = datetime.timedelta(minutes=15),
                                                      obs=[[-1.0,-2.0,-3.0],[-1.1,-2.1,-3.1],[-1.2,-2.2,-3.2]])
        mapnames = {'v[%d]'%i : s2.id+'[%d]'%i for i in range(3)}
        df2 = df2.rename(columns=mapnames)

        expected_df1 = pd.merge(df1, df2, how='outer', left_index=True, right_index=True)
        expected_df2 = pd.merge(df2, df1, how='outer', left_index=True, right_index=True)
        print('expected dataframe1')
        print(expected_df1)
        print('expected dataframe2')
        print(expected_df2)

        print('creating stream and observations %s' % s1.id)
        self.api.create_stream(s1)
        self.api.create_observations(o1, streamid=s1.id)

        print('creating stream and observations %s' % s2.id)
        self.api.create_stream(s2)
        self.api.create_observations(o2, streamid=s2.id)

        CSVparser = PandasObservationParser()
        print('getting observations stream ordering #1')
        retrieived_observations1 = self.api.get_observations(streamid='%s,%s' % (s1.id,s2.id), media='csv', parser=CSVparser)
        print(retrieived_observations1)
        print('getting observations stream ordering #2')
        retrieived_observations2 = self.api.get_observations(streamid='%s,%s' % (s2.id,s1.id), media='csv', parser=CSVparser)
        print(retrieived_observations2)

        assert_frame_equal(expected_df1, retrieived_observations1)
        assert_frame_equal(expected_df2, retrieived_observations2)

        self.api.destroy_observations(streamid=s1.id)
        self.api.destroy_stream(id=s1.id)
        self.api.destroy_observations(streamid=s2.id)
        self.api.destroy_stream(id=s2.id)

    @tape.use_cassette('test_pandas_get_observations_scalar_vector_stream.json')
    def test_pandas_get_observations_scalar_vector_stream(self):
        s1 = self.generate_vector_stream(3)
        o1, points1, df1 = self.generate_observations(basetimestamp = datetime.datetime(2016, 2, 15, 0, 0, 0, tzinfo=UTC_INFO),
                                                      deltat = datetime.timedelta(minutes=15),
                                                      obs=[[1.0,2.0,3.0],[1.1,2.1,3.1],[1.2,2.2,3.2]])
        mapnames = {'v[%d]'%i : s1.id+'[%d]'%i for i in range(3)}
        df1 = df1.rename(columns=mapnames)

        s2 = self.generate_scalar_stream()
        o2, points2, df2 = self.generate_observations(basetimestamp = datetime.datetime(2016, 2, 15, 0, 15, 0, tzinfo=UTC_INFO),
                                                      deltat = datetime.timedelta(minutes=15),
                                                      obs=[1.5,2.5,3.5])
        mapnames = {'v': s2.id}
        df2 = df2.rename(columns=mapnames)

        expected_df1 = pd.merge(df1, df2, how='outer', left_index=True, right_index=True)
        expected_df2 = pd.merge(df2, df1, how='outer', left_index=True, right_index=True)
        print('expected dataframe1')
        print(expected_df1)
        print('expected dataframe2')
        print(expected_df2)

        print('creating stream and observations %s' % s1.id)
        self.api.create_stream(s1)
        self.api.create_observations(o1, streamid=s1.id)

        print('creating stream and observations %s' % s2.id)
        self.api.create_stream(s2)
        self.api.create_observations(o2, streamid=s2.id)

        CSVparser = PandasObservationParser()
        print('getting observations stream ordering #1')
        retrieived_observations1 = self.api.get_observations(streamid='%s,%s' % (s1.id,s2.id), media='csv', parser=CSVparser)
        print(retrieived_observations1)
        print('getting observations stream ordering #2')
        retrieived_observations2 = self.api.get_observations(streamid='%s,%s' % (s2.id,s1.id), media='csv', parser=CSVparser)
        print(retrieived_observations2)

        assert_frame_equal(expected_df1, retrieived_observations1,check_column_type=CHECK_INFERRED_TYPES)
        assert_frame_equal(expected_df2, retrieived_observations2,check_column_type=CHECK_INFERRED_TYPES)

        self.api.destroy_observations(streamid=s1.id)
        self.api.destroy_stream(id=s1.id)
        self.api.destroy_observations(streamid=s2.id)
        self.api.destroy_stream(id=s2.id)

    def test_pandas_get_aggregation_scalar_stream(self):
        s = self.generate_scalar_stream()
        o, points, df = self.generate_observations(obs=[1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0, 11.0])

        expected_data = [[pd.to_datetime('2016-02-15 00:00:00+00:00'), 2.5, 1.0, 4.0, 4],
                         [pd.to_datetime('2016-02-15 01:00:00+00:00'), 6.5, 5.0, 8.0, 4],
                         [pd.to_datetime('2016-02-15 02:00:00+00:00'), 10.0, 9.0, 11.0, 3]]
        expected_columns = ['timestamp',s.id+'.avg', s.id+'.min', s.id+'.max', s.id+'.count']
        expected_df = pd.DataFrame(expected_data, columns=expected_columns)
        print('expected dataframe')
        print(expected_df)

        print('creating stream and observations %s' % s.id)
        self.api.create_stream(s)
        self.api.create_observations(o, streamid=s.id)

        print('getting observations')
        CSVparser = PandasObservationParser()
        retrieved_observations = self.api.get_aggregation(streamid=s.id, parser=CSVparser, aggperiod=1000*60*60)
        print(retrieved_observations)

        assert_frame_equal(expected_df, retrieved_observations)

        self.api.destroy_observations(streamid=s.id)
        self.api.destroy_stream(id=s.id)



"""
MIT License
Copyright (c) 2016 Ionata Digital

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""
from __future__ import absolute_import, print_function
from .config import SensorApiTestCase, username, tape

import unittest
import six
if six.PY3:
    from unittest.case import skip
else:
    from unittest2.case import skip

class AuthTestCase(SensorApiTestCase):
    @tape.use_cassette('test_credentials_configured.json')
    def test_credentials_configured(self):
        self.assertEqual(username, self.api.me.id,
                         "Expected user and actual user don't match.")

    @tape.use_cassette('test_roles.json')
    def test_roles(self):
        api_roles = self.api.me.roles
        assert len(api_roles) >= 1 #at least one role

    def test_auth_header_not_unicode(self):

        from requests import PreparedRequest

        from requests.structures import CaseInsensitiveDict

        p = PreparedRequest()
        p.headers = CaseInsensitiveDict()
        r = self.api.auth(p)

        k = r.headers.keys()
        for key in k:
            if key == 'Authorization':
                self.assertTrue(type(key) == str, 'Authorization header is unicode not str. Ensure unicode_literals are not imported.')



"""
MIT License
Copyright (c) 2016 Ionata Digital
Copyright (c) 2009-2014 Joshua Roesslein

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""
from requests.auth import _basic_auth_str


class AuthBase(object):
    """Base class that all auth implementations derive from"""

    def __call__(self, r):
        raise NotImplementedError('Auth hooks must be callable.')

    def get_username(self):
        raise NotImplementedError('Auth get_username() must be implemented.')


class HTTPBasicAuth(AuthBase):
    """Attaches HTTP Basic Authentication to the given Request object."""

    def __init__(self, username, password):
        self.username = username
        self.password = password

    def __call__(self, r):
        r.headers['Authorization'] = _basic_auth_str(self.username, self.password)
        return r

    def get_username(self):
        return self.username


class HTTPKeyAuth(AuthBase):
    """
    Attaches an API key to the given Request object.
    
    NOTE: as the corresponding username is unknown, using this auth scheme will
    break the api.me() method.
    
    TODO: figure out if there's a way to support API keys in query parameters
    instead of headers.
    """
    
    def __init__(self, key, header='apikey'):
        self.key = key
        self.header = header
    
    def __call__(self, r):
        r.headers[self.header] = self.key
        return r


class HTTPConsumerIDAuth(AuthBase):
    """
    Attaches an 'X-Consumer-Custom-ID' to the given Request object.

    Useful only for internal use, as this method is not valid outside the Senaps firewall.
    """

    def __init__(self, username):
        self.username = username

    def __call__(self, r):
        r.headers['X-Consumer-Custom-ID'] = self.username
        return r

    def get_username(self):
        return self.username



"""
MIT License
Copyright (c) 2016 Ionata Digital
Copyright (c) 2009-2014 Joshua Roesslein

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from __future__ import unicode_literals, absolute_import, print_function

import json

import datetime
import enum

from senaps_sensor.error import SenapsError
from senaps_sensor.utils import SenseTEncoder
from senaps_sensor.vocabulary import find_unit_of_measurement, find_observed_property


class StreamResultType(enum.Enum):
    scalar = "scalarvalue"
    geolocation = "geolocationvalue"
    vector = "vectorvalue"
    image = "imagevalue"
    document = "documentvalue"

class StreamMetaDataType(enum.Enum):
    scalar = ".ScalarStreamMetaData"
    geolocation = ".GeoLocationStreamMetaData"
    vector = ".VectorStreamMetaData"
    regularly_binned_vector = ".RegularlyBinnedVectorStreamMetaData"
    image = ".ImageStreamMetaData"
    document = ".DocumentStreamMetaData"

class StreamMetaDataMimeType(enum.Enum):
    json = "application/json"
    text = "text/plain"

class InterpolationType(enum.Enum):
    continuous = 'http://www.opengis.net/def/waterml/2.0/interpolationType/Continuous'
    discontinuous = 'http://www.opengis.net/def/waterml/2.0/interpolationType/Discontinuous'
    instant_total = 'http://www.opengis.net/def/waterml/2.0/interpolationType/InstantTotal'
    average_preceding = 'http://www.opengis.net/def/waterml/2.0/interpolationType/AveragePrec'
    max_preceding = 'http://www.opengis.net/def/waterml/2.0/interpolationType/MaxPrec'
    min_preceding = 'http://www.opengis.net/def/waterml/2.0/interpolationType/MinPrec'
    total_preceding = 'http://www.opengis.net/def/waterml/2.0/interpolationType/TotalPrec'
    const_preceding = 'http://www.opengis.net/def/waterml/2.0/interpolationType/ConstPrec'
    average_succeeding = 'http://www.opengis.net/def/waterml/2.0/interpolationType/AverageSucc'
    total_succeeding = 'http://www.opengis.net/def/waterml/2.0/interpolationType/TotalSucc'
    min_succeeding = 'http://www.opengis.net/def/waterml/2.0/interpolationType/MinSucc'
    max_succeeding = 'http://www.opengis.net/def/waterml/2.0/interpolationType/MaxSucc'
    const_succeeding = 'http://www.opengis.net/def/waterml/2.0/interpolationType/ConstSucc'

class ResultSet(list):
    """A list like object that holds results from a Twitter API query."""

    def __init__(self, max_id=None, since_id=None):
        super(ResultSet, self).__init__()
        self._max_id = max_id
        self._since_id = since_id

    @property
    def max_id(self):
        if self._max_id:
            return self._max_id
        ids = self.ids()
        # Max_id is always set to the *smallest* id, minus one, in the set
        return (min(ids) - 1) if ids else None

    @property
    def since_id(self):
        if self._since_id:
            return self._since_id
        ids = self.ids()
        # Since_id is always set to the *greatest* id in the set
        return max(ids) if ids else None

    def ids(self):
        return [item.id for item in self if hasattr(item, 'id')]


class Model(object):
    misspellings = {
        # key: wrong, value: correct
        'cummulative': 'cumulative',
    }

    def __init__(self, api=None):
        self._api = api
        self._fix_spellings = False

    def __getstate__(self, action=None):
        # pickle
        pickle = dict(self.__dict__)
        try:
            for key in [k for k in pickle.keys() if k.startswith('_')]:
                del pickle[key]  # do not pickle private attrs
        except KeyError:
            pass

        if self._fix_spellings:
            for wrong, correct in self.misspellings.items():
                if correct in pickle.keys():
                    pickle[wrong] = pickle.get(correct)
                    del pickle[correct]

        # allow model implementations to mangle state on different api actions
        action_fn = getattr(self, "__getstate_{0}__".format(action), None)
        if action and callable(action_fn):
            pickle = action_fn(pickle)

        return pickle

    def to_state(self, action=None):
        state = self.__getstate__(action)
        return state

    def to_json(self, action=None, indent=None):
        return json.dumps(self.to_state(action), sort_keys=True, cls=SenseTEncoder,
                          indent=indent)  # be explict with key order so unittest work.

    @classmethod
    def parse(cls, api, json_frag):
        instance = cls(api)

        try:
            instance.usermetadata = json_frag['usermetadata']
        except KeyError:
            pass

        return instance

    @classmethod
    def parse_list(cls, api, json_list):
        """
            Parse a list of JSON objects into
            a result set of model instances.
        """
        results = ResultSet()
        for obj in json_list:
            if obj:
                results.append(cls.parse(api, obj))
        return results

    @classmethod
    def fix_parse_misspellings(cls, json_frag):
        for wrong, correct in cls.misspellings.items():
            if wrong in json_frag.keys():
                json_frag[correct] = json_frag.get(wrong)
                del json_frag[wrong]

    def __repr__(self):
        state = ['%s=%s' % (k, repr(v)) for (k, v) in vars(self).items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(state))

    def __eq__(self, other):
        if self.id == other.id:
            return True
        else:
            return False


class JSONModel(Model):
    @classmethod
    def parse(cls, api, json_frag):
        return json_frag


class Platform(Model):
    def __init__(self, api=None):
        super(Platform, self).__init__(api=api)
        self._organisations = list()
        self._groups = list()
        self._streams = list()
        self._deployments = list()

    def __getstate__(self, action=None):
        pickled = super(Platform, self).__getstate__(action)

        pickled["groupids"] = [g.id for g in self.groups]
        pickled["streamids"] = [s.id for s in self.streams]
        pickled["deployments"] = [d.__getstate__(action) for d in self.deployments]
        return pickled

    def __getstate_create__(self, pickled):
        """
        :param pickled: dict of object kay, values
        :return: API weirdly requires a single organisationid on creation/update but returns a list
        """
        if not self.organisations:
            raise SenapsError("Platform creation requires an organisationid.")
        pickled["organisationid"] = self.organisations[0].id
        return pickled

    def __getstate_update__(self, pickled):
        """
        :param pickled: dict of object kay, values
        :return: pointer to self.__getstate_create__
        """
        return self.__getstate_create__(pickled)

    @classmethod
    def parse(cls, api, json_frag):
        platform = super(Platform, cls).parse(api, json_frag)
        setattr(platform, '_json', json_frag)
        for k, v in json_frag.items():
            if k == "_embedded":
                for ek, ev in v.items():
                    if ek == "organisation":
                        setattr(platform, "organisations", Organisation.parse_list(api, ev))
                    elif ek == "groups":
                        setattr(platform, "groups", Group.parse_list(api, ev))
                    elif ek == "platformdeployment":
                        setattr(platform, "deployments", Deployment.parse_list(api, ev))
                    elif ek == "streams":
                        setattr(platform, "streams", Stream.parse_list(api, ev))
            else:
                setattr(platform, k, v)
        return platform

    @classmethod
    def parse_list(cls, api, json_list):
        if isinstance(json_list, list):
            item_list = json_list
        elif "_embedded" not in json_list:
            item_list = []
        else:
            item_list = json_list['_embedded']['platforms']

        results = ResultSet()
        for obj in item_list:
            results.append(cls.parse(api, obj))
        return results

    @property
    def organisations(self):
        return self._organisations

    @organisations.setter
    def organisations(self, value):
        self._organisations = value

    @property
    def groups(self):
        return self._groups

    @groups.setter
    def groups(self, value):
        self._groups = value

    @property
    def streams(self):
        return self._streams

    @streams.setter
    def streams(self, value):
        self._streams = value

    @property
    def deployments(self):
        return self._deployments

    @deployments.setter
    def deployments(self, value):
        self._deployments = value


class Organisation(Model):
    @classmethod
    def parse(cls, api, json_frag):
        organisation = super(Organisation, cls).parse(api, json_frag)
        setattr(organisation, '_json', json_frag)
        for k, v in json_frag.items():
            setattr(organisation, k, v)
        return organisation

    @classmethod
    def parse_list(cls, api, json_list):
        if isinstance(json_list, list):
            item_list = json_list
        else:
            item_list = json_list['organisations']

        results = ResultSet()
        for obj in item_list:
            results.append(cls.parse(api, obj))
        return results

    def permissions(self):
        raise NotImplementedError("Not implemented.")


class Vocabulary(Model):
    pass


class StreamMetaData(Model):
    def __init__(self, api=None):
        super(StreamMetaData, self).__init__(api=api)
        self._fix_spellings = True
        self._type = None
        self._interpolation_type = None

        # document only attrs
        self._mimetype = None

        # scalar only attrs
        self._observed_property = None
        self.cumulative = None

        # geo only attrs
        self._unit_of_measure = None

        # simple vector only attrs
        self._length = None
        # regularly binned vector only attrs
        self._start = None
        self._end = None
        self._step = None
        self._amplitude_unit = None
        self._length_unit = None

        # cumulative stream only attrs
        self.accumulationInterval = None
        self.accumulationAnchor = None

        # required for cumulative scalar streams
        self.timezone = None

    def __getstate__(self, action=None):
        pickled = super(StreamMetaData, self).__getstate__(action)

        if self.interpolation_type:
            pickled["interpolationType"] = self.interpolation_type.value

        if self.observed_property:
            pickled["observedProperty"] = self.observed_property

        if self.unit_of_measure:
            pickled["unitOfMeasure"] = self.unit_of_measure

        if self.amplitude_unit:
            pickled["amplitudeUnit"] = self.amplitude_unit

        if self.length_unit:
            pickled["lengthUnit"] = self.length_unit

        # clean up non document StreamMetaData keys
        if self._type != StreamMetaDataType.document:
            for key in ['mimetype']:
                try:
                    del pickled[key]
                except KeyError:
                    pass

        # clean up non scalar StreamMetaData keys
        if self._type != StreamMetaDataType.scalar and self._type != StreamMetaDataType.regularly_binned_vector:
            for key in ['observedProperty', 'cumulative']:
                try:
                    del pickled[key]
                except KeyError:
                    pass

        # clean up non geo StreamMetaData keys
        if self._type != StreamMetaDataType.scalar:
            for key in ['unitOfMeasure']:
                try:
                    del pickled[key]
                except KeyError:
                    pass

        # clean up non cumulative stream StreamMetaData keys
        if not self.cumulative:
            for key in ['accumulationInterval', 'accumulationAnchor']:
                try:
                    del pickled[key]
                except KeyError:
                    pass
            if self.cumulative is None:  # different then false in PAI
                del pickled['cummulative']

        if self.timezone is None:
            del pickled["timezone"]

        if action != "create":
            # purge the type, it is never returned on get request
            try:
                del pickled['type']
            except KeyError:
                pass

        return pickled

    def __getstate_create__(self, pickled):
        """
        :param pickled: dict of object kay, values
        :return:
        """
        if not self.type:
            raise SenapsError("Stream creation requires an type.")
        if self.type is not None:
            pickled["type"] = self._type.value

        if self.mimetype is not None:
            pickled["mimetype"] = self._mimetype.value

        if self.length is not None:
            pickled["length"] = self._length

        if self.start is not None:
            pickled["start"] = self._start

        if self.end is not None:
            pickled["end"] = self._end

        if self.step is not None:
            pickled["step"] = self._step
        return pickled

    @classmethod
    def parse(cls, api, json_frag):
        stream_meta_data = super(StreamMetaData, cls).parse(api, json_frag)
        cls.fix_parse_misspellings(json_frag)

        setattr(stream_meta_data, '_json', json_frag)
        for k, v in json_frag.items():
            if k == "type":
                setattr(stream_meta_data, "type", StreamMetaDataType(v))
            elif k == "mimetype":
                setattr(stream_meta_data, "mimetype", StreamMetaDataMimeType(v))
            elif k == "_embedded":
                for ek, ev in v.items():
                    if ek == "interpolationType":
                        ev = ev[0].get('_links', {}).get('self', {}).get('href', )
                        setattr(stream_meta_data, "interpolation_type", InterpolationType(ev))
                    elif ek == "observedProperty":
                        ev = ev[0].get('_links', {}).get('self', {}).get('href', )
                        # Remove local vocab checks for now
                        setattr(stream_meta_data, "observed_property", ev)
                    elif ek == "unitOfMeasure":
                        ev = ev[0].get('_links', {}).get('self', {}).get('href', )
                        # Remove local vocab checks for now
                        setattr(stream_meta_data, "unit_of_measure", ev)
                    elif ek == "amplitudeUnit":
                        ev = ev[0].get('_links', {}).get('self', {}).get('href', )
                        # Remove local vocab checks for now
                        setattr(stream_meta_data, "amplitude_unit", ev)
                    elif ek == "lengthUnit":
                        ev = ev[0].get('_links', {}).get('self', {}).get('href', )
                        # Remove local vocab checks for now
                        setattr(stream_meta_data, "length_unit", ev)
                    else:
                        setattr(stream_meta_data, ek, ev)
                        print("parse: %s, %s" % (ek, ev))
            else:
                setattr(stream_meta_data, k, v)
        return stream_meta_data

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, value):
        self._type = value

    @property
    def mimetype(self):
        return self._mimetype

    @mimetype.setter
    def mimetype(self, value):
        self._mimetype = value

    @property
    def interpolation_type(self):
        return self._interpolation_type

    @interpolation_type.setter
    def interpolation_type(self, value):
        self._interpolation_type = value

    @property
    def observed_property(self):
        return self._observed_property

    @observed_property.setter
    def observed_property(self, value):
        self._observed_property = value

    @property
    def unit_of_measure(self):
        return self._unit_of_measure

    @unit_of_measure.setter
    def unit_of_measure(self, value):
        self._unit_of_measure = value

    # vector properties
    @property
    def length(self):
        return self._length

    @length.setter
    def length(self, value):
        self._length = value

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, value):
        self._start = value

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, value):
        self._end = value

    @property
    def step(self):
        return self._step

    @step.setter
    def step(self, value):
        self._step = value

    @property
    def amplitude_unit(self):
        return self._amplitude_unit

    @amplitude_unit.setter
    def amplitude_unit(self, value):
        self._amplitude_unit = value

    @property
    def length_unit(self):
        return self._length_unit

    @length_unit.setter
    def length_unit(self, value):
        self._length_unit = value


class Stream(Model):
    def __init__(self, api=None):
        super(Stream, self).__init__(api=api)
        self._result_type = None
        self._organisations = list()
        self._groups = list()
        self._metadata = None
        self._location = None

    def __getstate__(self, action=None):
        pickled = super(Stream, self).__getstate__(action)

        pickled["resulttype"] = self._result_type.value if self._result_type is not None else None
        pickled["organisationid"] = self.organisations[0].id

        if self.groups:
            pickled["groupids"] = [g.id for g in self.groups]

        try:
            if self.location:
                pickled["locationid"] = self.location.id
        except AttributeError:
            # excetion will be thrown if location is not specified and current object does not have a location
            pass

        if self.metadata:
            pickled["streamMetadata"] = self.metadata.__getstate__(action)

        return pickled

    @classmethod
    def parse(cls, api, json_frag):
        stream = super(Stream, cls).parse(api, json_frag)
        setattr(stream, '_json', json_frag)
        for k, v in json_frag.items():
            if k == "resulttype":
                setattr(stream, "result_type", StreamResultType(v))
            elif k == "_embedded":
                for ek, ev in v.items():
                    if ek == "organisation":
                        setattr(stream, "organisations", Organisation.parse_list(api, ev))
                    elif ek == "groups":
                        setattr(stream, "groups", Group.parse_list(api, ev))
                    elif ek == "location":
                        setattr(stream, "location", Location.parse(api, ev[0]))
                    elif ek == "metadata":
                        # metadata is also a list ?????
                        setattr(stream, "metadata", StreamMetaData.parse(api, ev[0]))
            else:
                setattr(stream, k, v)
        return stream

    @classmethod
    def parse_list(cls, api, json_list):
        if isinstance(json_list, list):
            item_list = json_list
        elif ('_embedded' in json_list
              and 'streams' in json_list['_embedded']):
            item_list = json_list['_embedded']['streams']
        elif ('count' in json_list
              and json_list['count'] == 0):
            item_list = {}
        else:
            raise SenapsError('Unable to parse list: [%s]' % ', '.join(map(str, json_list)))

        results = ResultSet()
        for obj in item_list:
            results.append(cls.parse(api, obj))
        return results

    @property
    def result_type(self):
        return self._result_type

    @result_type.setter
    def result_type(self, value):
        self._result_type = value

    @property
    def organisations(self):
        return self._organisations

    @organisations.setter
    def organisations(self, value):
        self._organisations = value

    @property
    def groups(self):
        return self._groups

    @groups.setter
    def groups(self, value):
        self._groups = value

    @property
    def location(self):
        return self._location

    @location.setter
    def location(self, value):
        self._location = value

    @property
    def metadata(self):
        return self._metadata

    @metadata.setter
    def metadata(self, value):
        self._metadata = value


class Group(Model):
    @classmethod
    def parse(cls, api, json_frag):
        group = super(Group, cls).parse(api, json_frag)
        setattr(group, '_json', json_frag)
        for k, v in json_frag.items():
            setattr(group, k, v)
        return group

    @classmethod
    def parse_list(cls, api, json_list):
        if isinstance(json_list, list):
            item_list = json_list
        else:
            item_list = json_list.get('_embedded', {}).get('groups', [])

        results = ResultSet()
        for obj in item_list:
            results.append(cls.parse(api, obj))
        return results


# TODO - not all attributes are implemented
class Location(Model):

    def __init__(self, api=None):
        super(Location, self).__init__(api=api)
        self._organisations = list()
        self._groups = list()

    @classmethod
    def parse(cls, api, json_frag):
        location = super(Location, cls).parse(api, json_frag)

        setattr(location, '_json', json_frag)
        for k, v in json_frag.items():
            if k == "_embedded":
                for ek, ev in v.items():
                    if ek == "organisation":
                        setattr(location, "organisations", Organisation.parse_list(api, ev))
                    elif ek == "groups":
                        setattr(location, "groups", Group.parse_list(api, ev))
            else:
                setattr(location, k, v)

        return location

    @classmethod
    def parse_list(cls, api, json_list):
        if isinstance(json_list, list):
            item_list = json_list
        elif "_embedded" not in json_list:
            item_list = []
        else:
            item_list = json_list['_embedded']['locations']

        results = ResultSet()
        for obj in item_list:
            results.append(cls.parse(api, obj))
        return results

    def __getstate__(self, action=None):
        pickled = super(Location, self).__getstate__(action)
        pickled["groupids"] = [g.id for g in self.groups]
        pickled["organisationid"] = self.organisations[0].id

        return pickled

    @property
    def organisations(self):
        return self._organisations

    @organisations.setter
    def organisations(self, value):
        self._organisations = value

    @property
    def groups(self):
        return self._groups

    @groups.setter
    def groups(self, value):
        self._groups = value


class Procedure(Model):
    pass


class Observation(Model):
    def __init__(self, api=None):
        super(Observation, self).__init__(api=api)
        self._results = list()
        self._stream = None

    def __getstate__(self, action=None):
        pickled = super(Observation, self).__getstate__(action)

        pickled["results"] = [r.to_state(action) for r in self.results] if self.results else []
        if self.stream:
            pickled["streamid"] = self.stream.to_state(action).get("id")
        return pickled

    @classmethod
    def parse(cls, api, json_frag):
        observation = super(Observation, cls).parse(api, json_frag)
        setattr(observation, '_json', json_frag)
        for k, v in json_frag.items():
            if k == "results":
                setattr(observation, "results", UnivariateResult.parse_list(api, v))
            if k == "stream":
                setattr(observation, "stream", Stream.parse(api, v))
            else:
                setattr(observation, k, v)
        return observation

    @classmethod
    def parse_list(cls, api, json_list):
        if isinstance(json_list, list):
            item_list = json_list
        else:
            item_list = json_list['observations']

        results = ResultSet()
        for obj in item_list:
            results.append(cls.parse(api, obj))
        return results

    @classmethod
    def from_dataframe(cls, dataframe):
        result = {}

        for timestamp, series in dataframe.iterrows():
            timestamp = timestamp.strftime('%Y-%m-%dT%H:%M:%S.%fZ')

            for series_id, value in series.iteritems():
                observation = UnivariateResult(t=timestamp, v=value)
                result.setdefault(series_id, Observation()).results.append(observation)
        return result

    @property
    def results(self):
        return self._results

    @results.setter
    def results(self, value):
        self._results = value

    @property
    def stream(self):
        return self._stream

    @stream.setter
    def stream(self, value):
        self._stream = value


class Aggregation(Model):
    def __init__(self, api=None):
        super(Aggregation, self).__init__(api=api)
        self._results = list()
        self._stream = None

    def __getstate__(self, action=None):
        pickled = super(Aggregation, self).__getstate__(action)

        pickled["results"] = [r.to_state(action) for r in self.results] if self.results else []
        if self.stream:
            pickled["streamid"] = self.stream.to_state(action).get("id")
        return pickled

    @classmethod
    def parse(cls, api, json_frag):
        aggregation = super(Aggregation, cls).parse(api, json_frag)
        setattr(aggregation, '_json', json_frag)
        for k, v in json_frag.items():
            if k == "results":
                setattr(aggregation, "results", UnivariateResult.parse_list(api, v))
            if k == "stream":
                setattr(aggregation, "stream", Stream.parse(api, v))
            else:
                setattr(aggregation, k, v)
        return aggregation

    @classmethod
    def parse_list(cls, api, json_list):
        if isinstance(json_list, list):
            item_list = json_list
        else:
            item_list = json_list['aggregations']

        results = ResultSet()
        for obj in item_list:
            results.append(cls.parse(api, obj))
        return results

    @classmethod
    def from_dataframe(cls, dataframe):
        result = {}

        for timestamp, series in dataframe.iterrows():
            timestamp = timestamp.strftime('%Y-%m-%dT%H:%M:%S.%fZ')

            for series_id, value in series.iteritems():
                aggregation = UnivariateResult(t=timestamp, v=value)
                result.setdefault(series_id, Observation()).results.append(aggregation)
        return result

class UnivariateResult(JSONModel):
    def __init__(self, api=None, t=None, v=None):
        super(UnivariateResult, self).__init__(api=api)
        self.t = t
        self.v = v

    def __getstate__(self, action=None):
        pickled = super(UnivariateResult, self).__getstate__(action)

        if isinstance(pickled.get('t', None), datetime.datetime):
            pickled['t'] = pickled.get('t').strftime('%Y-%m-%dT%H:%M:%S.%fZ')
        return pickled


class Deployment(Model):
    @classmethod
    def parse(cls, api, json_frag):
        deployment = super(Deployment, cls).parse(api, json_frag)
        setattr(deployment, '_json', json_frag)

        for k, v in json_frag.items():
            if k == "_embedded":
                for ek, ev in v.items():
                    if ek == "location":
                        setattr(deployment, "_location", Location.parse(api, ev[0]))
            else:
                setattr(deployment, k, v)

        return deployment

    @classmethod
    def parse_list(cls, api, json_list):
        if isinstance(json_list, list):
            item_list = json_list
        else:
            item_list = json_list['deployments']

        results = ResultSet()
        for obj in item_list:
            results.append(cls.parse(api, obj))
        return results

    def __getstate__(self, action=None):
        pickled = super(Deployment, self).__getstate__(action)

        if self.location:
            pickled["locationid"] = self.location.id

        return pickled

    @property
    def location(self):
        if hasattr(self,'_location'):
            return self._location

    @location.setter
    def location(self, value):
        self._location = value

    def permissions(self):
        raise NotImplementedError("Not implemented.")


class Role(Model):

    @classmethod
    def parse(cls, api, json_frag):
        role = super(Role, cls).parse(api, json_frag)
        setattr(role, '_json', json_frag)
        for k, v in json_frag.items():
            setattr(role, k, v)

        # look for org + groupid in _embedded data.
        # nb: these aren't always present, need to be careful with check.
        if '_embedded' not in json_frag:
            return role

        # admin, org, and group type roles, each comes with different embedded data.
        if 'group' in json_frag['_embedded']:
            setattr(role, 'groupid', json_frag['_embedded']['group'][0]['id'])
        if 'organisation' in json_frag['_embedded']:
            setattr(role, 'organisationid', json_frag['_embedded']['organisation'][0]['id'])

        return role

    @classmethod
    def parse_list(cls, api, json_list):
        if isinstance(json_list, list):
            item_list = json_list
        else:
            item_list = json_list['_embedded']['roles']

        results = ResultSet()
        for obj in item_list:
            results.append(cls.parse(api, obj))
        return results

    def permissions(self):
        raise NotImplementedError("Not implemented.")


class User(Model):

    @classmethod
    def parse(cls, api, json_frag):
        user = super(User, cls).parse(api, json_frag)
        attrs = [
            'id',
            'hidden',
            '_links',
            '_embedded',
        ]
        setattr(user, '_json', json_frag)
        for k, v in json_frag.items():
            if k in attrs:
                setattr(user, k, v)
        return user

    @classmethod
    def parse_list(cls, api, json_list):
        if isinstance(json_list, list):
            item_list = json_list
        else:
            item_list = json_list['_embedded']['users']

        results = ResultSet()
        for obj in item_list:
            results.append(cls.parse(api, obj))
        return results

    @property
    def roles(self):
        if hasattr(self, '_embedded') and 'roles' in self._embedded.keys():
            return Role.parse_list(self._api, self._embedded.get('roles'))
        return None

    def groups(self):
        pass


class Permitted(Model):

    @classmethod
    def parse(cls, api, json_frag):
        permitted = super(Permitted, cls).parse(api, json_frag)
        attrs = [
            '_links',
            '_embedded',
        ]
        setattr(permitted, '_json', json_frag)
        setattr(permitted, 'permitted', json_frag['permitted'])
        for k, v in json_frag.items():
            if k in attrs:
                setattr(permitted, k, v)

        return permitted

    @property
    def user(self):
        if hasattr(self, '_embedded') and 'user' in self._embedded.keys():
            return User.parse(self._api, self._embedded.get('user')[0])
        return None


class ModelFactory(object):
    """
    Used by parsers for creating instances
    of models. You may subclass this factory
    to add your own extended models.
    """
    user = User
    organisation = Organisation
    group = Group
    role = Role
    vocabulary = Vocabulary
    stream = Stream
    platform = Platform
    location = Location
    procedure = Procedure
    observation = Observation
    aggregation = Aggregation
    permitted = Permitted

    json = JSONModel



"""
MIT License
Copyright (c) 2016 Ionata Digital
Copyright (c) 2009-2014 Joshua Roesslein

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from __future__ import print_function, unicode_literals, absolute_import

import six
import re

from senaps_sensor.models import ModelFactory
from senaps_sensor.utils import import_simplejson
from senaps_sensor.error import SenapsError


if six.PY2:
    from cStringIO import StringIO
else:
    from io import StringIO as StringIO

class Parser(object):

    def parse(self, method, payload):
        """
        Parse the response payload and return the result.
        Returns a tuple that contains the result data and the cursors
        (or None if not present).
        """
        raise NotImplementedError

    def parse_error(self, payload):
        """
        Parse the error message and api error code from payload.
        Return them as an (error_msg, error_code) tuple. If unable to parse the
        message, throw an exception and default error message will be used.
        """
        raise NotImplementedError


class RawParser(Parser):

    def __init__(self):
        pass

    def parse(self, method, payload):
        return payload

    def parse_error(self, payload):
        return payload


class JSONParser(Parser):

    payload_format = 'json'

    def __init__(self):
        self.json_lib = import_simplejson()

    def parse(self, method, payload):
        try:
            json = self.json_lib.loads(payload)
        except Exception as e:
            raise SenapsError('Failed to parse JSON payload: %s' % e)

        needs_cursors = 'cursor' in method.session.params
        if needs_cursors and isinstance(json, dict):
            if 'previous_cursor' in json:
                if 'next_cursor' in json:
                    cursors = json['previous_cursor'], json['next_cursor']
                    return json, cursors
        else:
            return json

    def parse_error(self, payload):
        error_object = self.json_lib.loads(payload)
        reason = "An unknown error occurred"
        api_code = None

        if 'status' in error_object or 'message' in error_object:
            reason = error_object.get('message', reason)
            api_code = error_object.get('status', None)

        return reason, api_code


class ModelParser(JSONParser):

    def __init__(self, model_factory=None):
        JSONParser.__init__(self)
        self.model_factory = model_factory or ModelFactory

    def parse(self, method, payload):
        try:
            if method.payload_type is None:
                return
            model = getattr(self.model_factory, method.payload_type)
        except AttributeError:
            raise SenapsError('No model for this payload type: '
                             '%s' % method.payload_type)

        json = JSONParser.parse(self, method, payload)
        if isinstance(json, tuple):
            json, cursors = json
        else:
            cursors = None

        if method.payload_list:
            result = model.parse_list(method.api, json)
        else:
            result = model.parse(method.api, json)

        if cursors:
            return result, cursors
        else:
            return result

class PandasObservationParser(Parser):
    def __init__(self):
        import pandas # NOTE: import here means we don't require pandas to be installed unless we actually instantiate this class.
        self.pandas = pandas
        
        self.json_lib = import_simplejson()
    
    def parse(self, method, payload):
        # Validate media type.
        media_type = method.query_params.get('media', None)
        stream_ids = method.query_params['streamid'].split(',')  # NOTE: this WILL break if stream IDs contain commas (need to properly parse as CSV).
        aggperiod = method.query_params.get('aggperiod', None)
        if aggperiod is None:
            if media_type != 'csv':
                raise SenapsError('Observation query with PandasObservationParser requires CSV media type (media type "{}" is not supported).'.format(media_type))
            else:
                # Skip header information.
                lines = payload.splitlines()
                for i, row in enumerate(lines):
                    if 'timestamp' == row.split(',')[0]:
                        break

                # Parse CSV payload.
                df = self.pandas.read_csv(StringIO('\n'.join(lines[i:])), parse_dates=True, index_col='timestamp')

        else:
            # Parse json payload from Aggregation query.
            data = self.json_lib.loads(payload)
            if (self.pandas.__version__,) < ('1.0',):
                df = self.pandas.io.json.json_normalize(data['results'])
                # reorder columns
                df = df[['t', 'v.avg', 'v.min', 'v.max', 'v.count']]
            else:
                df = self.pandas.json_normalize(data['results'])
            sid = method.query_params['streamid']
            # rename to make it look like an Observation query
            df = df.rename(columns={'t': 'timestamp', 'v.avg': sid+'.avg', 'v.min': sid+'.min', 'v.max': sid+'.max', 'v.count': sid+'.count'})
            df['timestamp'] = self.pandas.to_datetime(df['timestamp'])
            df.set_index('timestamp')

        # Senaps returns columns in random (alphabetic?) order - reorder to
        # match the order the stream IDs were originally given in.
        if len(stream_ids) > 1:
            # cater for vectors which have multiple headers streamid[0], streamid[1]...
            df_headers = list(df)
            full_id_list = []
            for id in stream_ids:
                full_id_list.extend([col for col in df_headers if id == re.sub("[\[].*?[\]]", "", col)])
            df = df[full_id_list]

        return df
    
    def parse_error(self, payload):
        error_object = self.json_lib.loads(payload)
        reason = "An unknown error occurred"
        api_code = None
        
        if 'status' in error_object or 'message' in error_object:
            reason = error_object.get('message', reason)
            api_code = error_object.get('status', None)
        
        return reason, api_code



# Senaps data platform v2 API client library.
#
# Originally created by
# Copyright 2016 Ionata Digital
#
# Now maintained as part of Senaps platform by
# Copyright 2021 CSIRO
#
# See LICENSE for details.

"""
Senaps data platform sensor API client library
"""

from __future__ import absolute_import, unicode_literals, print_function
from .api import API

__version__ = "2.17.0"
__author__ = "CSIRO Data61"
__license__ = "MIT"

api = API()
